from typing import Dict

from model import Model
from utils import preprocess

from servicefoundry.function_service import remote


class ModelComposition:
    def __init__(self, model_1_path: str, model_2_path: str, model_3_path: str):
        # self.model_1 = Model(model_path="foo")
        # self.model_2 = Model(model_path="bar")
        # self.model_3 = Model(model_path="baz")
        self.model_1 = remote(
            Model, init_kwargs={"model_path": model_1_path}, name="model_1"
        )
        self.model_2 = remote(
            Model, init_kwargs={"model_path": model_2_path}, name="model_2"
        )
        self.model_3 = remote(
            Model, init_kwargs={"model_path": model_3_path}, name="model_3"
        )

    def predict(self, image_url: str) -> Dict:
        # preprocessed_data = preprocess(image_url)
        preprocessed_data = remote(preprocess)(image_url=image_url)

        model_1_prediction = self.model_1.predict(preprocessed_data=preprocessed_data)
        model_2_prediction = self.model_2.predict(preprocessed_data=preprocessed_data)
        model_3_prediction = self.model_3.predict(preprocessed_data=preprocessed_data)

        return dict(
            model_1_prediction=model_1_prediction,
            model_2_prediction=model_2_prediction,
            model_3_prediction=model_3_prediction,
        )


if __name__ == "__main__":

    model = ModelComposition(model_1_path="foo", model_2_path="bar", model_3_path="baz")
    print(model.predict(image_url="hello"))



import logging

import rich_click as click

from servicefoundry.cli.config import CliConfig
from servicefoundry.cli.const import COMMAND_CLS, GROUP_CLS
from servicefoundry.cli.display_util import print_json
from servicefoundry.cli.util import handle_exception_wrapper
from servicefoundry.lib.dao import application as application_lib
from servicefoundry.lib.dao import workspace as workspace_lib
from servicefoundry.logger import logger

# TODO (chiragjn): --json should disable all non json console prints


@click.group(name="delete", cls=GROUP_CLS)
def delete_command():
    """
    Delete Truefoundry resources

    \b
    Supported resources:
    - Workspace
    - Application
    """
    pass


@click.command(name="workspace", cls=COMMAND_CLS, help="Delete a Workspace")
@click.option(
    "-w",
    "--workspace-fqn",
    "--workspace_fqn",
    type=click.STRING,
    default=None,
    help="FQN of the Workspace to delete",
    required=True,
)
@click.confirmation_option(prompt="Are you sure you want to delete this workspace?")
@handle_exception_wrapper
def delete_workspace(workspace_fqn):
    deleted_workspace = workspace_lib.delete_workspace(
        workspace_fqn=workspace_fqn,
    )
    if CliConfig.get("json"):
        print_json(data=deleted_workspace.dict())


@click.command(name="application", cls=COMMAND_CLS, help="Delete an Application")
@click.option(
    "--application-fqn",
    "--application_fqn",
    type=click.STRING,
    default=None,
    help="FQN of the Application to delete",
    required=True,
)
@click.confirmation_option(prompt="Are you sure you want to delete this application?")
@handle_exception_wrapper
def delete_application(application_fqn):
    response = application_lib.delete_application(
        application_fqn=application_fqn,
    )
    if CliConfig.get("json"):
        print_json(data=response)


def get_delete_command():
    delete_command.add_command(delete_workspace)
    delete_command.add_command(delete_application)
    return delete_command



from __future__ import annotations

from typing import Optional

from servicefoundry.lib.auth.credential_provider import (
    CredentialProvider,
    EnvCredentialProvider,
    FileCredentialProvider,
)
from servicefoundry.lib.model.entity import UserInfo
from servicefoundry.logger import logger

ACTIVE_SESSION: Optional[ServiceFoundrySession] = None


class ServiceFoundrySession:
    def __init__(self) -> None:
        self._cred_provider = self._get_cred_provider()
        self._user_info: UserInfo = self._cred_provider.token.to_user_info()

        global ACTIVE_SESSION
        if (ACTIVE_SESSION is None) or (
            ACTIVE_SESSION
            and ACTIVE_SESSION.base_url != self.base_url
            and ACTIVE_SESSION.user_info != self.user_info
        ):
            logger.info(
                "Logged in to %r as %r (%s)",
                self.base_url,
                self.user_info.user_id,
                self.user_info.email or self.user_info.user_type.value,
            )
        ACTIVE_SESSION = self

    @staticmethod
    def _get_cred_provider() -> CredentialProvider:
        final_cred_provider = None
        for cred_provider in [EnvCredentialProvider, FileCredentialProvider]:
            if cred_provider.can_provide():
                final_cred_provider = cred_provider()
                break
        if final_cred_provider is None:
            raise Exception(
                "Please login again using `sfy login --relogin`"
                "or `sfy.login(relogin=True)` function"
            )
        return final_cred_provider

    @property
    def access_token(self) -> str:
        return self._cred_provider.token.access_token

    @property
    def base_url(self) -> str:
        return self._cred_provider.base_url

    @property
    def user_info(self) -> UserInfo:
        return self._user_info



from __future__ import annotations

import datetime
import time
from enum import Enum
from typing import Any, Dict, List, Optional, Union

import jwt
from pydantic import BaseModel, Extra, Field, constr, validator

# TODO: switch to Enums for str literals
# TODO: Need a better approach to keep fields in sync with server
#       most fields should have a default in case server adds/removes a field
# TODO: Implement NotImplementedError sections
UNDEFINED_STRING = "<Undefined>"


class Base(BaseModel):
    class Config:
        validate_assignment = True
        use_enum_values = True
        extra = Extra.allow

    def __repr_args__(self):
        return [
            (key, value)
            for key, value in self.__dict__.items()
            if key in self.__fields__
            and self.__fields__[key].field_info.extra.get("repr", True)
        ]


class Entity(Base):
    createdAt: datetime.datetime = Field(repr=False)
    updatedAt: datetime.datetime = Field(repr=False)
    createdBy: str = Field(repr=False)


class Cluster(Entity):
    id: str = Field(repr=False)
    name: str
    fqn: str
    region: str = Field(repr=False)
    isTenantDefault: bool = False
    isAvailable: bool = False

    def list_row_data(self) -> Dict[str, Any]:
        return {
            "ClusterName": self.name,
            "Region": self.region,
            "CreatedAt": self.createdAt,
        }

    def get_data(self) -> Dict[str, Any]:
        return {
            "ClusterName": self.name,
            "Region": self.region,
            "CreatedAt": self.createdAt,
            "UpdatedAt": self.updatedAt,
            "CreatedBy": self.createdBy,
        }

    @property
    def workspaces(self) -> List["Workspace"]:
        raise NotImplementedError


class Workspace(Entity):
    id: str = Field(repr=False)
    name: str
    fqn: str
    clusterId: str = Field(repr=False)

    def list_row_data(self) -> Dict[str, Any]:
        return {
            "WorkspaceName": self.name,
            "WorkspaceFQN": self.fqn,
            "ClusterName": self.clusterId,
            "CreatedAt": self.createdAt,
        }

    def get_data(self) -> Dict[str, Any]:
        return {
            "WorkspaceName": self.name,
            "WorkspaceFQN": self.fqn,
            "ClusterName": self.clusterId,
            "CreatedAt": self.createdAt,
            "UpdatedAt": self.updatedAt,
            "CreatedBy": self.createdBy,
        }

    @property
    def cluster(self) -> Cluster:
        raise NotImplementedError


class WorkspaceResources(BaseModel):
    cpu_limit: Optional[float]
    memory_limit: Optional[int]
    ephemeral_storage_limit: Optional[int]


class PortMetadata(BaseModel):
    port: int
    # TODO: done because of a bug, needs to reverted to be mandatory
    host: Optional[str]


class DeploymentTransitionStatus(str, Enum):
    INITIALIZED: str = "INITIALIZED"
    BUILDING: str = "BUILDING"
    DEPLOYING: str = "DEPLOYING"
    BUILD_SUCCESS: str = "BUILD_SUCCESS"
    DEPLOY_SUCCESS: str = "DEPLOY_SUCCESS"
    DEPLOY_FAILED: str = "DEPLOY_FAILED"
    BUILD_FAILED: str = "BUILD_FAILED"
    CANCELLED: str = "CANCELLED"
    FAILED: str = "FAILED"
    REUSING_EXISTING_BUILD: str = "REUSING_EXISTING_BUILD"
    _: str = ""

    @classmethod
    def is_failure_state(cls, state: DeploymentTransitionStatus) -> bool:
        return state in (cls.DEPLOY_FAILED, cls.BUILD_FAILED, cls.FAILED, cls.CANCELLED)


class DeploymentState(BaseModel):
    isTerminalState: bool


class DeploymentStatus(BaseModel):
    state: DeploymentState
    status: DeploymentTransitionStatus
    transition: Optional[DeploymentTransitionStatus]


class DeploymentManifest(Base):
    name: Optional[str] = Field(default=UNDEFINED_STRING)
    type: str


class Deployment(Entity):
    id: str = Field(repr=False)
    fqn: str
    version: int
    currentStatusId: str = Field(repr=False)
    applicationId: str = Field(repr=False)
    manifest: DeploymentManifest = Field(repr=False)
    failureReason: Optional[str]
    deploymentStatuses: Optional[List[DeploymentStatus]]
    # TODO: Dict -> pydantic model if required
    # application: Dict[str, Any]
    # TODO: Dict -> pydantic model if required
    # workspace: Dict[str, Any]
    # baseDomainURL: str
    # builds: List[BuildResponse]
    metadata: Optional[
        Union[Dict[str, Any], List[Dict[str, Any]]]
    ]  # TODO (chiragjn): revisit the type of this field
    currentStatus: Optional[DeploymentStatus]  # Server was not returning CurrentStatus

    def list_row_data(self) -> Dict[str, Any]:
        return {
            "DeploymentFQN": self.fqn,
            "ApplicationName": self.manifest.name,
            "ApplicationType": self.manifest.type,
            "Version": self.version,
            "CreatedAt": self.createdAt,
        }

    def get_data(self) -> Dict[str, Any]:
        return {
            "DeploymentFQN": self.fqn,
            "ApplicationName": self.manifest.name,
            "ApplicationType": self.manifest.type,
            "Version": self.version,
            "WorkspaceName": self.fqn.split(":")[1],  # TODO: change hardcoding
            "ClusterName": self.fqn.split(":")[0],  # TODO: change hardcoding
            "CreatedAt": self.createdAt,
            "UpdatedAt": self.updatedAt,
            "CreatedBy": self.createdBy,
        }

    class Config:
        extra = Extra.allow


class ApplicationWorkspace(Base):
    name: str
    fqn: str
    clusterId: str


class Application(Entity):
    id: str = Field(repr=False)
    name: str
    fqn: str
    tenantName: str
    workspaceId: str = Field(repr=False)
    lastVersion: int
    activeVersion: int
    workspace: ApplicationWorkspace
    deployment: Deployment
    activeDeploymentId: Optional[str] = Field(repr=False)
    lastDeploymentId: str = Field(repr=False)

    def list_row_data(self) -> Dict[str, Any]:
        return {
            "ApplicationName": self.name,
            "ApplicationFQN": self.fqn,
            "ActiveVersion": self.activeVersion,
            "LastVersion": self.lastVersion,
            "WorkspaceName": self.workspace.name,
            "CreatedAt": self.createdAt,
        }

    def get_data(self) -> Dict[str, Any]:
        return {
            "ApplicationName": self.name,
            "ApplicationFQN": self.fqn,
            "ActiveVersion": self.activeVersion,
            "LastVersion": self.lastVersion,
            "WorkspaceFQN": self.workspace.fqn,
            "ClusterFQN": self.workspace.clusterId,
            "CreatedAt": self.createdAt,
            "UpdatedAt": self.updatedAt,
            "CreatedBy": self.createdBy,
        }


class UserType(Enum):
    user = "user"
    serviceaccount = "serviceaccount"


class UserInfo(BaseModel):
    user_id: constr(min_length=1)
    user_type: UserType = UserType.user
    email: Optional[str] = None
    tenant_name: constr(min_length=1) = Field(alias="tenantName")

    class Config:
        allow_population_by_field_name = True
        allow_mutation = False


class TenantInfo(BaseModel):
    tenant_name: constr(min_length=1) = Field(alias="tenantName")
    auth_server_url: str

    class Config:
        allow_population_by_field_name = True
        allow_mutation = False


class Token(BaseModel):
    access_token: constr(min_length=1) = Field(alias="accessToken", repr=False)
    refresh_token: Optional[constr(min_length=1)] = Field(
        alias="refreshToken", repr=False
    )
    decoded_value: Optional[Dict] = Field(exclude=True, repr=False)

    class Config:
        allow_population_by_field_name = True
        allow_mutation = False

    @validator("decoded_value", always=True, pre=True)
    def _decode_jwt(cls, v, values, **kwargs):
        access_token = values["access_token"]
        return jwt.decode(
            access_token,
            options={
                "verify_signature": False,
                "verify_aud": False,
                "verify_exp": False,
            },
        )

    @property
    def tenant_name(self) -> str:
        return self.decoded_value["tenantName"]

    def is_going_to_be_expired(self, buffer_in_seconds: int = 120) -> bool:
        exp = int(self.decoded_value["exp"])
        return (exp - time.time()) < buffer_in_seconds

    def to_user_info(self) -> UserInfo:
        return UserInfo(
            user_id=self.decoded_value["username"],
            email=self.decoded_value["email"]
            if "email" in self.decoded_value
            else None,
            user_type=UserType(self.decoded_value.get("userType", UserType.user.value)),
            tenant_name=self.tenant_name,
        )


class CredentialsFileContent(BaseModel):
    access_token: constr(min_length=1) = Field(repr=False)
    refresh_token: Optional[constr(min_length=1)] = Field(repr=False)
    host: constr(min_length=1)

    class Config:
        allow_mutation = False

    def to_token(self) -> Token:
        return Token(access_token=self.access_token, refresh_token=self.refresh_token)


class DeviceCode(BaseModel):
    user_code: str = Field(alias="userCode")
    device_code: str = Field(alias="deviceCode")

    class Config:
        allow_population_by_field_name = True
        allow_mutation = False

    def get_user_clickable_url(self, auth_host: str) -> str:
        return f"{auth_host}/authorize/device?userCode={self.user_code}"


class TriggerJobResult(Base):
    message: str = Field(default="Unknown")


class DockerRegistryCredentials(Base):
    fqn: str
    registryUrl: str
    username: str
    password: str


class CreateDockerRepositoryResponse(Base):
    repoName: str



import importlib
import os
from threading import Thread

import uvicorn
from fastapi import FastAPI

from servicefoundry.function_service.remote import QUAL_NAME_TO_URL_MAP
from servicefoundry.function_service.route import ClassRoute, FunctionRoute, RouteGroups
from servicefoundry.function_service.utils import (
    create_pydantic_model_from_function_signature,
)


def _encapsulate_user_function(function, input_model_name: str):
    model = create_pydantic_model_from_function_signature(
        func=function, model_name=input_model_name
    )

    def func(x: model):
        return function(**dict(x._iter(to_dict=False)))

    return func


def _add_function_route(app: FastAPI, function_route: FunctionRoute, port: int):
    module = importlib.import_module(function_route.module)
    function = getattr(module, function_route.function_name)

    app.add_api_route(
        path=function_route.path,
        endpoint=_encapsulate_user_function(
            function, input_model_name=function_route.qual_name
        ),
        methods=[function_route.http_method],
    )
    QUAL_NAME_TO_URL_MAP[
        function_route.qual_name
    ] = f"http://localhost:{port}{function_route.path}"


def _add_class_route(app: FastAPI, class_route: ClassRoute, port: int):
    module = importlib.import_module(class_route.module)
    class_factory = getattr(module, class_route.class_name)
    instance = class_factory(**class_route.init_kwargs)

    for route in class_route.routes:
        function = getattr(instance, route.function_name)
        app.add_api_route(
            path=route.path,
            endpoint=_encapsulate_user_function(function, route.qual_name),
            methods=[route.http_method],
        )

        QUAL_NAME_TO_URL_MAP[route.qual_name] = f"http://localhost:{port}{route.path}"


def build_app(route_groups: RouteGroups, port: int) -> FastAPI:
    _root_path = os.getenv("TFY_SERVICE_ROOT_PATH")
    app_kwargs = {"root_path": _root_path} if _root_path else {}
    app = FastAPI(docs_url="/", **app_kwargs)
    app.add_api_route(path="/ping", endpoint=lambda: "pong")

    for route in route_groups.functions:
        _add_function_route(app=app, function_route=route, port=port)

    for class_route in route_groups.classes.values():
        _add_class_route(app=app, class_route=class_route, port=port)

    # print(json.dumps(QUAL_NAME_TO_URL_MAP, indent=2))
    return app


def build_and_run_app(route_groups: RouteGroups, port: int):
    app = build_app(route_groups=route_groups, port=port)
    uvicorn.run(app, port=port, host="0.0.0.0")


def build_and_run_app_in_background_thread(
    route_groups: RouteGroups, port: int
) -> Thread:
    thread = Thread(
        target=build_and_run_app,
        kwargs=dict(route_groups=route_groups, port=port),
        daemon=True,
    )
    thread.start()
    return thread



import os
from typing import Optional

import rich_click as click

from servicefoundry.io.output_callback import OutputCallBack
from servicefoundry.lib.auth.auth_service_client import AuthServiceClient
from servicefoundry.lib.auth.credential_file_manager import CredentialsFileManager
from servicefoundry.lib.auth.credential_provider import EnvCredentialProvider
from servicefoundry.lib.clients.service_foundry_client import (
    ServiceFoundryServiceClient,
)
from servicefoundry.lib.clients.utils import resolve_base_url
from servicefoundry.lib.const import (
    API_KEY_ENV_NAME,
    HOST_ENV_NAME,
    OLD_SFY_PROFILES_FILEPATH,
    OLD_SFY_SESSIONS_FILEPATH,
    RICH_OUTPUT_CALLBACK,
)
from servicefoundry.lib.messages import (
    PROMPT_ALREADY_LOGGED_OUT,
    PROMPT_LOGOUT_SUCCESSFUL,
)
from servicefoundry.lib.model.entity import CredentialsFileContent, Token
from servicefoundry.logger import logger

if OLD_SFY_PROFILES_FILEPATH.exists():
    logger.warning(
        "%s file is deprecated. You can delete this file now.",
        OLD_SFY_PROFILES_FILEPATH,
    )


if OLD_SFY_SESSIONS_FILEPATH.exists():
    logger.warning(
        "%s file is deprecated. You can delete this file now.",
        OLD_SFY_SESSIONS_FILEPATH,
    )


def login(
    api_key: Optional[str] = None,
    host: Optional[str] = None,
    relogin: bool = False,
    output_hook: OutputCallBack = RICH_OUTPUT_CALLBACK,
) -> bool:
    if API_KEY_ENV_NAME in os.environ and HOST_ENV_NAME in os.environ:
        logger.warning(
            "Skipping login because environment variables %s and "
            "%s are set and will be used when running servicefoundry. "
            "If you want to relogin then unset these environment keys.",
            HOST_ENV_NAME,
            API_KEY_ENV_NAME,
        )
        return False

    if EnvCredentialProvider.can_provide():
        logger.warning(
            "TFY_API_KEY env var is already set. "
            "When running servicefoundry, it will use the api key to authorize.\n"
            "Login will just save the credentials on disk."
        )

    host = resolve_base_url(host).strip("/")

    with CredentialsFileManager() as cred_file:
        if not relogin and cred_file.exists():
            cred_file_content = cred_file.read()
            if host != cred_file_content.host:
                if click.confirm(
                    f"Already logged in to {cred_file_content.host!r}\n"
                    f"Do you want to relogin to {host!r}?"
                ):
                    return login(api_key=api_key, host=host, relogin=True)

            user_info = cred_file_content.to_token().to_user_info()
            user_name_display_info = user_info.email or user_info.user_type.value
            print(
                f"Already logged in to {cred_file_content.host!r} as "
                f"{user_info.user_id!r} ({user_name_display_info})\n"
                "Please use `sfy login --relogin` or `sfy.login(relogin=True)` "
                "to force relogin"
            )
            return False

        if api_key:
            service_foundry_client = ServiceFoundryServiceClient(
                init_session=False, base_url=host
            )
            token = _login_with_api_key(
                api_key=api_key, service_foundry_client=service_foundry_client
            )
        else:
            auth_service = AuthServiceClient(base_url=host)
            # interactive login
            token = _login_with_device_code(base_url=host, auth_service=auth_service)

        cred_file_content = CredentialsFileContent(
            access_token=token.access_token,
            refresh_token=token.refresh_token,
            host=host,
        )
        cred_file.write(cred_file_content)

    user_info = token.to_user_info()
    user_name_display_info = user_info.email or user_info.user_type.value
    output_hook.print_line(
        f"Successfully logged in to {cred_file_content.host!r} as "
        f"{user_info.user_id!r} ({user_name_display_info})"
    )
    return True


def logout(
    output_hook: OutputCallBack = RICH_OUTPUT_CALLBACK,
) -> None:
    with CredentialsFileManager() as cred_file:
        if cred_file.delete():
            output_hook.print_line(PROMPT_LOGOUT_SUCCESSFUL)
        else:
            output_hook.print_line(PROMPT_ALREADY_LOGGED_OUT)


def _login_with_api_key(
    api_key: str, service_foundry_client: ServiceFoundryServiceClient
) -> Token:
    logger.debug("Logging in with api key")
    return service_foundry_client.get_token_from_api_key(api_key=api_key)


def _login_with_device_code(
    base_url: str,
    auth_service: AuthServiceClient,
    output_hook: OutputCallBack = RICH_OUTPUT_CALLBACK,
) -> Token:
    logger.debug("Logging in with device code")
    device_code = auth_service.get_device_code()
    url_to_go = device_code.get_user_clickable_url(auth_host=base_url)
    output_hook.print_line(f"Opening:- {url_to_go}")
    output_hook.print(
        "Please click on the above link if it is not "
        "automatically opened in a browser window."
    )
    click.launch(url_to_go)
    return auth_service.get_token_from_device_code(device_code=device_code.device_code)



import logging

from servicefoundry import Build, Job, PythonBuild, Schedule

logging.basicConfig(level=logging.INFO)

job = Job(
    name="my-job-manual",
    image=Build(
        build_spec=PythonBuild(command="python main.py --upto 30"),
    ),
)
job.deploy(workspace_fqn="v1:local:my-ws-2")


job = Job(
    name="my-job-sch",
    image=Build(
        build_spec=PythonBuild(command="python main.py --upto 30"),
    ),
    trigger=Schedule(schedule="*/60 * * * *"),
)
job.deploy(workspace_fqn="v1:local:my-ws-2")



import inspect
from typing import Any

import pydantic


def get_qual_name(obj):
    return f"{obj.__module__}.{obj.__qualname__}"


def create_pydantic_model_from_function_signature(func, model_name: str):
    # https://github.com/pydantic/pydantic/issues/1391
    (
        args,
        _,
        varkw,
        defaults,
        kwonlyargs,
        kwonlydefaults,
        annotations,
    ) = inspect.getfullargspec(func)
    defaults = defaults or []
    args = args or []
    if len(args) > 0 and args[0] == "self":
        del args[0]

    non_default_args = len(args) - len(defaults)
    defaults = [
        ...,
    ] * non_default_args + defaults

    keyword_only_params = {
        param: kwonlydefaults.get(param, Any) for param in kwonlyargs
    }
    params = {
        param: (annotations.get(param, Any), default)
        for param, default in zip(args, defaults)
    }

    class Config:
        extra = "allow"

    # Allow extra params if there is a **kwargs parameter in the function signature
    config = Config if varkw else None

    return pydantic.create_model(
        model_name,
        **params,
        **keyword_only_params,
        __base__=pydantic.BaseModel,
        __config__=config,
    )



from enum import Enum
from typing import Any, Dict, Union

from pydantic import BaseModel

default = Union[str, int, float, None]


class Kind(str, Enum):
    STRING = "string"
    NUMBER = "number"
    INTEGER = NUMBER
    FLOAT = "float"
    OPTIONS = "options"
    WORKSPACE = "tfy-workspace"
    FILE = "tfy-file"
    PYTHON_FILE = "tfy-python-file"


class Parameter(BaseModel):
    prompt: str
    default: default
    suggest: default


class OptionsParameter(Parameter):
    options: Dict[Any, str]


class WorkspaceParameter(Parameter):
    pass


def get_parameter(parameter: Dict):
    raise RuntimeError("Don't use")



from servicefoundry.cli.commands.build_command import get_build_command
from servicefoundry.cli.commands.build_logs_command import get_build_logs_command
from servicefoundry.cli.commands.create_command import get_create_command
from servicefoundry.cli.commands.delete_command import get_delete_command
from servicefoundry.cli.commands.deploy_v2_command import deploy_v2_command
from servicefoundry.cli.commands.get_command import get_get_command
from servicefoundry.cli.commands.list_command import get_list_command
from servicefoundry.cli.commands.login_command import get_login_command
from servicefoundry.cli.commands.logout_command import get_logout_command
from servicefoundry.cli.commands.logs_command import get_logs_command
from servicefoundry.cli.commands.patch_command import get_patch_command
from servicefoundry.cli.commands.redeploy_command import get_redeploy_command
from servicefoundry.cli.commands.trigger_command import get_trigger_command



import json
import shlex
from typing import Any, Dict, Optional, Sequence, Union

import servicefoundry.lib.dao.version as version_lib
from servicefoundry.cli.console import console
from servicefoundry.lib.clients.service_foundry_client import (
    ServiceFoundryServiceClient,
)
from servicefoundry.lib.model.entity import Deployment, TriggerJobResult
from servicefoundry.lib.util import get_application_fqn_from_deployment_fqn
from servicefoundry.logger import logger


def list_applications(
    application_type: str,
    workspace_fqn: Optional[str] = None,
    client: Optional[ServiceFoundryServiceClient] = None,
):
    client = client or ServiceFoundryServiceClient()
    if workspace_fqn:
        workspace = client.get_id_from_fqn(fqn=workspace_fqn, fqn_type="workspace")
        applications = client.list_applications(workspace_id=workspace["workspaceId"])
    else:
        applications = client.list_applications()

    if application_type != "all":
        applications = [
            application
            for application in applications
            if application.deployment.manifest.type == application_type
        ]
    return applications


def get_application(
    application_fqn: str,
    client: Optional[ServiceFoundryServiceClient] = None,
):
    client = client or ServiceFoundryServiceClient()
    application = client.get_id_from_fqn(fqn=application_fqn, fqn_type="app")
    application = client.get_application_info(
        application_id=application["applicationId"]
    )
    return application


def delete_application(
    application_fqn: str,
    client: Optional[ServiceFoundryServiceClient] = None,
) -> Dict[str, Any]:
    client = client or ServiceFoundryServiceClient()
    application = client.get_id_from_fqn(fqn=application_fqn, fqn_type="app")
    response = client.remove_application(application_id=application["applicationId"])

    console.print("""[yellow]Deleted Application[/]""")
    return response


def redeploy_application(
    application_fqn: str,
    version: int,
    wait: bool,
    client: Optional[ServiceFoundryServiceClient] = None,
) -> Deployment:
    from servicefoundry.v2.lib.deployable_patched_models import Application

    client = client or ServiceFoundryServiceClient()

    deployment_info = version_lib.get_version(
        application_fqn=application_fqn, version=version
    )

    manifest = deployment_info.manifest.dict()

    application_id = deployment_info.applicationId
    application_info = client.get_application_info(application_id=application_id)
    workspace_fqn = application_info.workspace.fqn

    application = Application.parse_obj(manifest)
    deployment = application.deploy(workspace_fqn=workspace_fqn, wait=wait)
    return deployment


def trigger_job(
    application_fqn: str,
    command: Optional[Union[str, Sequence[str]]] = None,
    params: Optional[Dict[str, str]] = None,
) -> TriggerJobResult:
    """
    Trigger a Job on Truefoundry platform

    Args:
        application_fqn: Fully Qualified Name of the Deployed Job (without the version number)
        command: command to run the job with, defaults to `None`. Can be a `str` or `List[str]`
            When `None`, the job is triggered with configured command at the time of deployment.
            When passed as a list, the command will be joined using `shlex.join`

    Returns:
        TriggerJobResult: metadata returning status of job trigger
    """
    client = ServiceFoundryServiceClient()

    if params and command:
        raise Exception("Either pass `command` or `params` but not both")

    try:
        # If user is passing in deployment fqn copied from UI, till we change the fqns on UI
        application_fqn = get_application_fqn_from_deployment_fqn(application_fqn)
        logger.warning(
            "Detected version number in `application_fqn`. "
            "Automatically discarding the version number. "
            f"This automatic conversion will be removed in future. "
            f"Please pass {application_fqn!r} as the value."
        )
    except ValueError:
        pass
    _application_info = client.get_application_info_by_fqn(
        application_fqn=application_fqn
    )
    application_info = client.get_application_info(
        application_id=_application_info.applicationId
    )
    if command:
        if not isinstance(command, str):
            command_str = shlex.join(command).strip()
        else:
            command_str = command.strip()
        message = f"Job triggered with command {command_str!r}"
    elif not params and not command:
        command_str = ""
        message = "Job triggered with pre-configured command"

    if params:

        # Check if params has any non string values
        for key, value in params.items():
            if not isinstance(value, str):
                raise Exception(
                    f"Invalid value {value!r} for key {key!r}. "
                    "Only string values are allowed for `params`"
                )
        command_str = ""
        message = f"Job triggered with params {params!r}"

    result = client.trigger_job(
        deployment_id=application_info.activeDeploymentId,
        component_name=application_info.name,
        command=command_str if command_str else None,
        params=params if params else None,
    )
    previous_runs_url = f"{client.base_url.strip('/')}/deployments/{application_info.id}?tab=previousRuns"
    logger.info(
        f"{message}.\n"
        f"You can check the status of your job run at {previous_runs_url}"
    )
    return result



import logging

import rich_click as click

from servicefoundry.cli.config import CliConfig
from servicefoundry.cli.console import console
from servicefoundry.cli.const import COMMAND_CLS, GROUP_CLS
from servicefoundry.cli.display_util import print_entity_list, print_json
from servicefoundry.cli.util import handle_exception_wrapper
from servicefoundry.lib.dao import application as application_lib
from servicefoundry.lib.dao import version as version_lib
from servicefoundry.lib.dao import workspace as workspace_lib
from servicefoundry.lib.messages import (
    PROMPT_NO_APPLICATIONS,
    PROMPT_NO_VERSIONS,
    PROMPT_NO_WORKSPACES,
)
from servicefoundry.logger import logger

# TODO (chiragjn): --json should disable all non json console prints


@click.group(name="list", cls=GROUP_CLS)
def list_command():
    # TODO (chiragjn): Figure out a way to update supported resources based on ENABLE_* flags
    """
    List Truefoundry resources

    \b
    Supported resources:
    - Workspace
    - Application
    - Application Version
    """


@click.command(name="workspaces", cls=COMMAND_CLS, help="List Workspaces")
@click.option(
    "-c",
    "--cluster-name",
    "--cluster_name",
    type=click.STRING,
    default=None,
    help="Cluster name",
)
@handle_exception_wrapper
def list_workspaces(cluster_name):
    workspaces = workspace_lib.list_workspaces(
        cluster_name=cluster_name,
    )
    if not workspaces:
        console.print(PROMPT_NO_WORKSPACES)
    else:
        workspaces.sort(key=lambda s: s.createdAt, reverse=True)

    if CliConfig.get("json"):
        print_json(data=workspaces)
    else:
        print_entity_list(
            "Workspaces",
            workspaces,
        )


@click.command(name="applications", cls=COMMAND_CLS, help="List Applications")
@click.option(
    "-w",
    "--workspace-fqn",
    "--workspace_fqn",
    type=click.STRING,
    default=None,
    help="FQN of the Workspace",
)
@click.option(
    "--application-type",
    "--application_type",
    type=click.Choice(["service", "job", "model-deployment", "all"]),
    default="all",
    help="Application Type",
)
@handle_exception_wrapper
def list_applications(workspace_fqn, application_type):
    applications = application_lib.list_applications(
        workspace_fqn=workspace_fqn, application_type=application_type
    )
    if not applications:
        console.print(PROMPT_NO_APPLICATIONS)
    else:
        applications.sort(key=lambda s: s.createdAt, reverse=True)
    if CliConfig.get("json"):
        print_json(data=applications)
    else:
        # TODO (chiragjn): Display columns here need to show workspace and cluster name!
        print_entity_list(
            "Applications",
            applications,
        )


@click.command(
    name="application-versions", cls=COMMAND_CLS, help="List Application Versions"
)
@click.option(
    "--application-fqn",
    "--application_fqn",
    type=click.STRING,
    default=None,
    help="FQN of the application",
    required=True,
)
@handle_exception_wrapper
def list_versions(application_fqn):
    versions = version_lib.list_versions(
        application_fqn=application_fqn,
    )
    if not versions:
        console.print(PROMPT_NO_VERSIONS)
    else:
        versions.sort(key=lambda s: s.createdAt, reverse=True)
    if CliConfig.get("json"):
        print_json(data=versions)
    else:
        # TODO (chiragjn): Display columns here need to show workspace and cluster name!
        print_entity_list(
            "Application Versions",
            versions,
        )


def get_list_command():
    list_command.add_command(list_workspaces)
    list_command.add_command(list_applications)
    list_command.add_command(list_versions)

    return list_command



from my_module import Model, preprocess

from servicefoundry.function_service import FunctionService, remote

preprocess_service = FunctionService("preprocess", port=4000)
preprocess_service.register_function(preprocess)

model_service = FunctionService("model-service", port=4001)
model_service.register_class(Model, init_kwargs={"path": "foo"}, name="model")

preprocess_service.run()
model_service.run().join()



import datetime
from typing import List, Optional

from pydantic import BaseModel, Extra, create_model


class BuildResponse(BaseModel):
    id: str
    name: str
    # TODO: make status an enum
    status: str
    # TODO: should we just make these fields
    # snake-case and add camelCase aliases?
    deploymentId: str
    componentName: str
    createdAt: datetime.datetime
    updatedAt: datetime.datetime
    imageUri: Optional[str]
    failureReason: Optional[str]
    getLogsUrl: str
    tailLogsUrl: str
    logsStartTs: int

    class Config:
        extra = Extra.allow


class AppDeploymentStatusResponse(BaseModel):
    state: create_model(
        "State",
        isTerminalState=(bool, ...),
        type=(str, ...),
        transitions=(List[str], ...),
    )

    id: str
    status: str
    message: Optional[str]
    transition: Optional[str]

    class Config:
        extra = Extra.allow


class DeploymentFqnResponse(BaseModel):
    deploymentId: str
    applicationId: str
    workspaceId: str


class ApplicationFqnResponse(BaseModel):
    applicationId: str
    workspaceId: str



from rich.console import Console
from rich.highlighter import ReprHighlighter
from rich.panel import Panel
from rich.text import Text

from servicefoundry.io.output_callback import OutputCallBack


def _text(line):
    return Text.from_ansi(str(line)) if "\x1b" in line else str(line)


class RichOutputCallBack(OutputCallBack):
    console = Console(soft_wrap=True)
    highlighter = ReprHighlighter()

    def print_header(self, line):
        self.console.rule(_text(line), style="cyan")

    def print_line(self, line):
        self.console.print(_text(line))

    def print_lines_in_panel(self, lines, header=None):
        self.console.print(Panel(self.highlighter("\n".join(lines)), title=header))

    def print_code_in_panel(self, lines, header=None, lang="python"):
        self.print_lines_in_panel(lines, header)



import json
import os

from servicefoundry.lib.binarydownloader import BinaryDependencies, BinaryName
from servicefoundry.lib.clients.shell_client import Shell


class Terragrunt:
    def __init__(self) -> None:
        self.terragrunt_output_cache = {}

    def apply_all(self, dir):
        print(
            Shell().execute_shell_command(
                [
                    BinaryDependencies().which(BinaryName.TERRAGRUNT),
                    "run-all",
                    "apply",
                    "--terragrunt-non-interactive",
                    f"--terragrunt-working-dir={dir}",
                    f"--terragrunt-tfpath={BinaryDependencies().which(BinaryName.TERRAFORM)}",
                ]
            )
        )

    def fetch_terragrunt_output(self, dir, path):
        if dir in self.terragrunt_output_cache:
            return self.terragrunt_output_cache[dir][path]["value"]

        current_dir = os.getcwd()
        os.chdir(dir)
        v = Shell().execute_shell_command(
            [
                BinaryDependencies().which(BinaryName.TERRAGRUNT),
                "output",
                f"--terragrunt-working-dir={dir}",
                f"--terragrunt-tfpath={BinaryDependencies().which(BinaryName.TERRAFORM)}",
                "--json",
            ]
        )
        os.chdir(current_dir)

        self.terragrunt_output_cache[dir] = json.loads(v)
        return self.terragrunt_output_cache[dir][path]["value"]



class CliConfig:
    __conf = {"json": False}

    @staticmethod
    def get(name):
        return CliConfig.__conf[name]

    @staticmethod
    def set(name, value):
        CliConfig.__conf[name] = value



import logging

import rich_click as click

from servicefoundry.cli.const import COMMAND_CLS
from servicefoundry.cli.util import handle_exception_wrapper
from servicefoundry.io.rich_output_callback import RichOutputCallBack
from servicefoundry.lib.session import logout
from servicefoundry.logger import logger


@click.command(name="logout", cls=COMMAND_CLS)
@handle_exception_wrapper
def logout_command():
    """
    Logout from current Truefoundry session
    """
    callback = RichOutputCallBack()
    logout(output_hook=callback)


def get_logout_command():
    return logout_command



import argparse
import logging

from servicefoundry import Build, PythonBuild, Resources, Service

logging.basicConfig(level=logging.INFO)

parser = argparse.ArgumentParser()
parser.add_argument("--workspace-fqn", type=str, required=True)
args = parser.parse_args()

service = Service(
    # name="my-service",
    name="my-service-1",
    image=Build(
        build_spec=PythonBuild(
            command="uvicorn main:app --port 4000 --host 0.0.0.0",
            pip_packages=[
                "fastapi",
                "uvicorn",
            ],
            python_version="3.9",
        ),
    ),
    ports=[{"expose": True, "port": 4000}],
    replicas=2,
    resources=Resources(cpu_limit=0.6),
)
deployment = service.deploy(workspace_fqn=args.workspace_fqn)



import rich_click as click
import yaml

from servicefoundry.cli.const import GROUP_CLS

# from servicefoundry.cli.display_util import print_json


@click.group(
    name="deploy",
    cls=GROUP_CLS,
    invoke_without_command=True,
    help="Deploy application to Truefoundry",
)
@click.option(
    "-f",
    "--file",
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    default="./servicefoundry.yaml",
    help="Path to servicefoundry.yaml file",
    show_default=True,
)
@click.option(
    "-w",
    "--workspace-fqn",
    "--workspace_fqn",
    type=click.STRING,
    required=True,
    help="FQN of the Workspace to deploy to",
)
@click.option(
    "--wait/--no-wait",
    "--wait/--no_wait",
    is_flag=True,
    show_default=True,
    default=True,
    help="Wait and tail the deployment progress",
)
def deploy_v2_command(file: str, workspace_fqn: str, wait: bool):
    from servicefoundry.v2.lib.deployable_patched_models import Application

    with open(file, "r") as f:
        application_definition = yaml.safe_load(f)

    application = Application.parse_obj(application_definition)
    _deployment = application.deploy(workspace_fqn=workspace_fqn, wait=wait)
    # print_json(_deployment)



from IPython.display import HTML, display
from pygments import highlight
from pygments.formatters import HtmlFormatter
from pygments.lexers import PythonLexer


def print_python_code(source_code):
    formatter = HtmlFormatter(noclasses=True)
    display(HTML((highlight(source_code, PythonLexer(), formatter))))



from __future__ import annotations

import os
from typing import Optional

import git
from pydantic import BaseModel

from servicefoundry.logger import logger


class GitRepo(BaseModel):
    # The git repository URL
    repo_url: str
    # Can be branch or git tag
    git_ref: str
    # Git username
    username: Optional[str]
    # Git password
    password: Optional[str]


class GitClient:
    def __init__(self) -> None:
        # TODO: Don't think this will be cross platform
        self.client = git.Git(os.path.join(os.path.expanduser("~"), "git", "GitPython"))

    def execute_git_command(self, args):
        try:
            result = self.client.execute(["git", *args])
            logger.info(result)
            return result
        except Exception as e:
            raise Exception("Git command failed") from e

    # Clone the repo and return the path
    def clone_repo(self, git_repo: GitRepo, destination_dir):
        input_repo_url: str = git_repo.repo_url
        if input_repo_url.startswith("http://"):
            raise Exception("Please provide a https git repository")
        if input_repo_url.startswith("https://"):
            input_repo_url = input_repo_url[8:]
        print("Git clone link: " + f"https://{git_repo.username}@{input_repo_url}")
        if git_repo.username and git_repo.password:
            self.execute_git_command(
                [
                    "clone",
                    f"https://{git_repo.username}:{git_repo.password}@{input_repo_url}",
                    destination_dir,
                ]
            )
        else:
            self.execute_git_command(
                [
                    "clone",
                    f"{git_repo.repo_url}",
                    destination_dir,
                ]
            )

    def commit_all_changes(self, git_repo_dir, branch):
        if branch is None:
            branch = "main"

        current_dir = os.getcwd()
        os.chdir(git_repo_dir)
        # Switch to the input branch
        self.execute_git_command(["checkout", "-B", branch])

        # Add all local files
        self.execute_git_command(["add", "."])
        diff_check = self.execute_git_command(["diff", "--staged"])
        if not diff_check:
            print(f"Nothing to commit in target repo: {git_repo_dir}")
        else:
            self.execute_git_command(
                ["commit", "-m", "K8s Cluster Configuration via ArgoCD applications"]
            )
            self.execute_git_command(["push", "-f", "origin", branch])

        os.chdir(current_dir)



import json
import os
import random
import shutil
import string
import tempfile
from typing import Optional

import chevron
import questionary
import yaml
from pydantic import BaseModel

from servicefoundry.lib.clients.cookiecutter_client import CookieCutter
from servicefoundry.lib.clients.git_client import GitClient, GitRepo
from servicefoundry.lib.clients.service_foundry_client import (
    ServiceFoundryServiceClient,
)
from servicefoundry.lib.clients.terragrunt_client import Terragrunt
from servicefoundry.lib.infra.utils import (
    check_if_crd_installed,
    check_if_truefoundry_exists,
    check_or_install_dependencies,
    create_namespace_if_not_exists,
    delete_directory,
    execute_kubectl_apply_file,
    execute_nats_bootstrap_script,
    install_argocd_chart,
    install_ubermold_helm,
    set_current_kubecontext,
)
from servicefoundry.logger import logger

THIS_DIR = os.path.dirname(os.path.abspath(__file__))


class HelmRepo(BaseModel):
    # Helm repo name
    name: str
    # The helm repository URL
    url: str
    # Token to access the helm repo
    token: Optional[str]


class InfraController:
    __infra_config = {}
    terragrunt_client = None
    helm_client = None

    def __init__(self, dry_run, config_file_path):
        if config_file_path:
            with open(config_file_path, "r") as config:
                self.__infra_config = yaml.safe_load(config)
        else:
            raise Exception("Input config file needs to be provided")
        self.__git_client = GitClient()
        self.dry_run = dry_run

    # Ask questions to user to compose the input infra config
    def ask_infra_config_from_user(self):
        logger.info("Let's get the infra setup for Truefoundry")
        return {}

    def __apply_terragrunt(self, base_terragrunt_dir):
        if not questionary.confirm(
            "Do you want to continue with infra creation: ", default=True
        ).ask():
            return
        # TODO A terragrunt repo which can cache the outputs for each module. This would reduce the number of output calls
        self.terragrunt_client = Terragrunt()
        self.terragrunt_client.apply_all(base_terragrunt_dir)

    def __add_ubermold_argocd_app(self, target_repo: GitRepo, target_repo_path: str):
        logger.info("====== Installing ArgoCD manifests========")
        if not questionary.confirm(
            f"""Have you already merged the commit to the main branch in {target_repo.repo_url}.
            Please make sure that is done before proceeding here. """,
            default=True,
        ).ask():
            return

        logger.info(
            f"""We will bootstrap ArgoCD to sync all the configuration present
            at {target_repo.repo_url}/{target_repo_path} to the
            Kubernetes cluster"""
        )

        input_argo_file = os.path.join(THIS_DIR, "argo-ubermold.mustache")
        output_argo_file = "argo-bootstrap.yaml"
        with open(input_argo_file, "r") as input_f, open(
            output_argo_file, "w"
        ) as output_f:
            output_f.write(
                chevron.render(
                    input_f,
                    {
                        "target_repo_url": target_repo.repo_url,
                        "target_repo_username": target_repo.username,
                        "target_repo_password": target_repo.password,
                        "target_repo_path": target_repo_path,
                    },
                )
            )
        # Apply the argo configuration to Kubernetes
        logger.info("Applying the argocd app to Kubernetes")
        execute_kubectl_apply_file(output_argo_file)
        logger.info(f"ArgoCD application spec at {output_argo_file} applied to cluster")

    # This function clones the target repo and return the clone_dir path
    def __clone_target_repo(self, target_repo: GitRepo):
        target_repo_clone_dir = os.path.join(tempfile.mkdtemp(), "target-repo")
        logger.info(
            f"""We will clone the {target_repo.repo_url} at {target_repo_clone_dir}.
            Please make sure you have permissions to clone and push to the repository
            from your local machine."""
        )
        # We will need to remove any credentials associated with the repo
        # since we assume that local machine has the credentials to push to that repo.
        target_repo_without_creds = GitRepo(
            repo_url=target_repo.repo_url,
            git_ref=target_repo.git_ref,
            username="",
            password="",
        )
        GitClient().clone_repo(target_repo_without_creds, target_repo_clone_dir)
        logger.info(f"Cloned target repo at {target_repo_clone_dir}")
        return target_repo_clone_dir

    def __get_local_truefoundry_config(self):
        return {
            "postgresql": {
                "auth": {
                    "existingSecret": "servicefoundry-server-env-secret",
                    "secretKeys": {
                        "adminPasswordKey": "DB_PASSWORD",
                        "userPasswordKey": "DB_PASSWORD",
                    },
                    "username": "truefoundry",
                    "database": "truefoundry",
                }
            }
        }

    def create_infra_using_terraform(self):
        logger.info(
            """
            We will install the Truefoundry infrastructure on the provided cloud account.
            Truefoundry can be installed on AWS, GCP or Azure cloud infra.
            The steps below will guide you through the installation process. If you are
            confused about any of the steps, please refer to docs at
            https://docs.truefoundry.com/docs/deploy-on-own-cloud-overview or contact us on
            Slack at https://truefoundry.slack.com/signup#/domain-signup
            """
        )

        # We have the input infra config. We use the input infra_config to get terraform inputs from server.
        response = ServiceFoundryServiceClient().process_infra(self.__infra_config)

        # Inputs for Terraform code
        terraform_inputs = response["terraform_inputs"]
        # Truefoundry template repo credentials
        template_git_repo = GitRepo.parse_obj(response["template_git_repo"])
        # Target Repo where to commit the terraform code
        target_repo = GitRepo(
            repo_url=response["target_repo"]["url"],
            git_ref=response["target_repo"]["branch"],
            username=response["target_repo"]["username"],
            password=response["target_repo"]["password"],
        )
        target_repo_path = response["target_repo"]["terraform_path"]

        # Clone the template code repo
        logger.info(
            "Cloning truefoundry template repo. git needs to be installed on the local system"
        )
        check_or_install_dependencies(["git"])
        clone_dir = os.path.join(tempfile.mkdtemp(), "tfy-repo")

        self.__git_client.clone_repo(template_git_repo, clone_dir)
        logger.info(f"Template repo cloned at: {clone_dir}")

        # Create cookiecutter.json file and render cookiecutter template
        cookiecutter_dir = os.path.join("infrastructure", terraform_inputs["provider"])

        cookiecutter_client = CookieCutter(clone_dir, directory=cookiecutter_dir)
        with open(
            os.path.join(clone_dir, cookiecutter_dir, "cookiecutter.json"), "w"
        ) as file:
            file.write(json.dumps(terraform_inputs, indent=2))
        rendered_code_path = cookiecutter_client.run(destination_dir=clone_dir)
        logger.info("The terraform code can be found at: " + rendered_code_path)

        # apply terragrunt
        tf_env_path = os.path.join(
            rendered_code_path,
            terraform_inputs["subscription"]["name"],
            terraform_inputs["location"]["name"],
            terraform_inputs["env"]["cluster_prefix"],
        )
        self.__apply_terragrunt(tf_env_path)
        logger.info(f"Terragrunt infra provisioning done")

        # We are skipping the step of generating the kubeconfig. We will ask the user
        # to do it manually for now - and assume that the kubeconfig is present at the
        # user's machine. Later we can do the kubeconfig generation maybe in the terraform
        # layer

        logger.info(f"""Infrastructure has been provisioned successfully""")
        # TODO: Test the below part
        # Clone target repo
        target_repo_clone_dir = os.path.join(tempfile.mkdtemp(), "target-repo")
        self.__git_client.clone_repo(target_repo, target_repo_clone_dir)
        logger.info(f"Cloned target repo at {target_repo_clone_dir}")

        # TODO: Log all relevant terragrunt outputs

        # Copy the terraform code folder into the path of target repo
        shutil.copytree(
            rendered_code_path,
            os.path.join(target_repo_clone_dir, target_repo_path),
        )
        # Push the content to Github
        GitClient().commit_all_changes(target_repo_clone_dir, target_repo.git_ref)

    def __commit_k8s_configuration_code_to_git(
        self, k8s_config_path: str, target_repo: GitRepo, target_repo_path: str
    ) -> bool:
        """
        This function commits the kubernetes configuration code to a git repository
        """
        logger.info(
            f"""The Kubernetes configuration is saved at path: {k8s_config_path}. You can commit
            the configuration to a Git repository and configure ArgoCD to sync from the Git repo.
            We will commit the code to the repo: {target_repo.repo_url} on the branch: {target_repo.git_ref}.
            You can then raise a PR and merge the code to the main branch after reviewing the code."""
        )
        if not questionary.confirm(
            "Are you ready to push the code to Git?",
            default=True,
        ).ask():
            return False
        # Clone target repo
        target_repo_clone_dir = self.__clone_target_repo(target_repo)

        # Copy the k8s configuration folder into the path of target repo
        shutil.copytree(
            os.path.join(k8s_config_path),
            os.path.join(target_repo_clone_dir, target_repo_path),
            dirs_exist_ok=True,
        )
        # Push the content to Github
        GitClient().commit_all_changes(target_repo_clone_dir, target_repo.git_ref)

        logger.info(
            f"""We have committed the changes to {target_repo.repo_url} in the branch: {target_repo.git_ref}.
            Please merge that branch to main branch"""
        )
        return True

    def setup_tfy_control_plane(
        self,
        db_host: str,
        db_username: str,
        db_password: str,
        db_name: str,
        tfy_svc_account_api_key: str,
        image_pull_secret: str,
    ):
        logger.info(
            "==== Installing Truefoundry control plane in the truefoundry namespace ===="
        )
        truefoundry_exists = check_if_truefoundry_exists()
        if truefoundry_exists:
            if not questionary.confirm(
                """Truefoundry is already installed on the current cluster. Do you want to override the current installation?
                    This can lead to loss of data for the current installation.""",
                default=False,
            ).ask():
                return
        logger.info("Create namespace truefoundry")
        create_namespace_if_not_exists("truefoundry")

        # Execute nats bootstrap script and get the seed
        nats_controlplane_account_seed = execute_nats_bootstrap_script()

        input_tfy_file = os.path.join(THIS_DIR, "tfy-control-plane.mustache")
        output_tfy_file = "tfy-control-plane.yaml"
        with open(input_tfy_file, "r") as input_f, open(
            output_tfy_file, "w"
        ) as output_f:
            output_f.write(
                chevron.render(
                    input_f,
                    {
                        "truefoundry_db_host": db_host,
                        "truefoundry_db_username": db_username,
                        "truefoundry_db_password": db_password,
                        "truefoundry_db_name": db_name,
                        "nats_controlplane_account_seed": nats_controlplane_account_seed,
                        "truefoundry_svc_account_api_key": tfy_svc_account_api_key,
                        "image_pull_secret": image_pull_secret,
                    },
                )
            )
        logger.info("Create the secrets in the truefoundry namespace")
        execute_kubectl_apply_file(output_tfy_file)
        delete_directory("./nsc")
        logger.info(
            f"Created secrets in the truefoundry namespace with values at {output_tfy_file}"
        )
        return

    def __setup_tfy_agent(self, cluster_token: str, image_pull_secret: str):
        logger.info("====== Installing agent in namespace: tfy-agent ======")
        create_namespace_if_not_exists("tfy-agent")

        input_file = os.path.join(THIS_DIR, "tfy-agent.mustache")
        output_file = "tfy-agent.yaml"
        with open(input_file, "r") as input_f, open(output_file, "w") as output_f:
            output_f.write(
                chevron.render(
                    input_f,
                    {
                        "image_pull_secret": image_pull_secret,
                        "cluster_token": cluster_token,
                    },
                )
            )
        logger.info("Setting up secrets for agent in tfy-agent namespace")
        execute_kubectl_apply_file(output_file)
        logger.info(
            f"Created secrets in the tfy-agent namespace with values at {output_file}"
        )

    def __check_or_install_tfy_base(self, tfy_helm_repo: HelmRepo) -> bool:
        """
        This function checks if the cluster is ready for Truefoundry installation. The key
        primary dependencies to install Truefoundry are argocd and istio. If they are
        not already installed, this function will guide the user to get those
        bootstrapped on the cluster.
        """

        logger.info(
            """We will need kubectl and helm to be installed on local system to be able
            to install truefoundry on the Kubernetes cluster"""
        )
        check_or_install_dependencies(["kubectl", "helm"])

        logger.info(
            "Set the current kubecontext to the cluster where you want to install Truefoundry"
        )
        set_current_kubecontext()

        # If istio is already present, ask user if they want to move ahead with their own istio configuration
        if check_if_crd_installed("istio", "virtualservices.networking.istio.io"):
            logger.info(
                """"Istio is already installed on this cluster.
                        If istio was installed via Truefoundry previously, then you can go ahead - else
                        Truefoundry's istio installation can wipe out your existing Istio configuration."""
            )
            if not questionary.confirm(
                "Are you ready to go ahead with overwriting existing istio installation: ",
                default=False,
            ).ask():
                return False

        if check_if_crd_installed("prometheus", "prometheuses.monitoring.coreos.com"):
            logger.info(
                """"Prometheus is already installed on this cluster.
                If prometheus was installed via Truefoundry previously, then you can go ahead - else
                Truefoundry's istio installation will wipe out your existing Prometheus configuration."""
            )
            if not questionary.confirm(
                "Are you ready to go ahead with overwriting Prometheus installation: ",
                default=False,
            ).ask():
                return False

        # Install ArgoCD in the cluster if its not already installed
        if not check_if_crd_installed("argocd", "applications.argoproj.io"):
            install_argocd_chart("5.16.13")

        # Add argo secret to access truefoundry helm charts repo
        input_argo_file = os.path.join(THIS_DIR, "argo-secrets.mustache")
        output_argo_file = "argo-secrets.yaml"
        with open(input_argo_file, "r") as input_f, open(
            output_argo_file, "w"
        ) as output_f:
            output_f.write(
                chevron.render(
                    input_f,
                    {
                        "tfy_helm_repo_url": tfy_helm_repo.url,
                        "tfy_helm_repo_token": tfy_helm_repo.token,
                    },
                )
            )
        # Apply the argo configuration to Kubernetes
        logger.info(
            "Creating secrets in argocd namespace to access truefoundry helm-charts"
        )
        execute_kubectl_apply_file(output_argo_file)
        logger.info(f"Secrets spec at {output_argo_file} applied to cluster")

        return True

    def __render_kubernetes_configuration(
        self,
        template_git_repo: GitRepo,
        cloud_provider: str,
        tenant_name: str,
        control_plane_url: str,
        install_control_plane: bool,
        install_agent: bool,
        local: bool,
    ):
        # Clone tfy-template_repo template and overwrite supermold values
        clone_dir = os.path.join(tempfile.mkdtemp(), "tfy-repo")
        logger.info(f"Cloning template repo at {clone_dir}")
        self.__git_client.clone_repo(template_git_repo, clone_dir)

        # Ask the user if they want to change the cookiecutter configuration
        cookiecutter_config_path = os.path.join(clone_dir, "k8s", "cookiecutter.json")
        config_dict = None
        # Read the cookiecutter config
        with open(cookiecutter_config_path, "r") as f:
            config_dict = json.load(f)
        # Modify the values in the cookiecutter config based on the inputs
        project_slug = config_dict["project_slug"]
        config_dict["tenantName"] = tenant_name
        config_dict["controlPlaneURL"] = control_plane_url
        config_dict["provider"] = cloud_provider
        if install_control_plane:
            config_dict["truefoundry"]["enabled"] = True
            if local:
                config_dict["truefoundry"][
                    "local"
                ] = self.__get_local_truefoundry_config()
                config_dict["truefoundry"]["create_vs"] = False
        if install_agent:
            config_dict["tfyAgent"]["enabled"] = True

        with open(cookiecutter_config_path, "w") as f:
            f.write(json.dumps(config_dict, indent=2))

        # Render the cookiecutter configuration
        cookiecutter_client = CookieCutter(clone_dir, directory="k8s")
        rendered_code_path = cookiecutter_client.run(
            destination_dir=os.path.join(clone_dir, "k8s")
        )
        target_path = os.path.join(clone_dir, "k8s", project_slug)
        logger.info(f"The kubernetes code can be found at: {target_path}")
        return target_path

    def install_truefoundry_to_k8s(self, local):
        logger.info("==== Bootstrapping the Kubernetes cluster with truefoundry ======")

        logger.info(
            """To bootstrap truefoundry (either control-plane or agent), we need to have argocd and
            istio installed on the cluster. If they are not already installed, we will guide you with the
            installation process as you proceed. If you are confused about any of the steps, please
            refer to docs at https://docs.truefoundry.com/docs/deploy-on-own-cloud-overview or contact us on
            Slack at https://truefoundry.slack.com/signup#/domain-signup"""
        )

        response = ServiceFoundryServiceClient().process_infra(self.__infra_config)
        template_git_repo = GitRepo.parse_obj(response["template_git_repo"])
        tfy_helm_repo = HelmRepo.parse_obj(response["tfy_helm_repo"])
        if "image_pull_secret" not in response.keys():
            logger.error(
                """Image Pull Credentials not provided in server response.
                Please contact truefoundry team to resolve this issue"""
            )
            return
        image_pull_secret = response["image_pull_secret"]
        truefoundry_svc_account_api_key = response["sfy_user_api_key"]
        tenant_name = response["tenant_name"]
        control_plane_url = self.__infra_config["control_plane_url"]
        cloud_provider = self.__infra_config["provider"]

        # Make sure cluster has the prerequisites for truefoundry installation
        if not self.__check_or_install_tfy_base(tfy_helm_repo):
            return
        # Check what all we need to install in the cluster
        install_tfy_control_plane = False
        if "tfy_control_plane" in self.__infra_config.keys():
            install_tfy_control_plane = True
            if local:
                db_password = str("".join(random.choices(string.ascii_letters, k=10)))
                self.setup_tfy_control_plane(
                    "truefoundry-postgresql.truefoundry.svc.cluster.local",
                    "truefoundry",
                    db_password,
                    "truefoundry",
                    truefoundry_svc_account_api_key,
                    image_pull_secret,
                )
            else:
                self.setup_tfy_control_plane(
                    self.__infra_config["tfy_control_plane"]["params"][
                        "truefoundry_db_host"
                    ],
                    self.__infra_config["tfy_control_plane"]["params"][
                        "truefoundry_db_username"
                    ],
                    self.__infra_config["tfy_control_plane"]["params"][
                        "truefoundry_db_password"
                    ],
                    self.__infra_config["tfy_control_plane"]["params"][
                        "truefoundry_db_name"
                    ],
                    truefoundry_svc_account_api_key,
                    image_pull_secret,
                )
        # Setup tfy_control_plane
        install_tfy_agent = False
        if "tfy_agent" in self.__infra_config.keys():
            install_tfy_agent = True
            self.__setup_tfy_agent(
                self.__infra_config["tfy_agent"]["cluster_token"], image_pull_secret
            )

        k8s_configuration_path = self.__render_kubernetes_configuration(
            template_git_repo,
            cloud_provider,
            tenant_name,
            control_plane_url,
            install_tfy_control_plane,
            install_tfy_agent,
            local,
        )

        if "target_repo" not in response.keys():
            install_ubermold_helm(k8s_configuration_path)
            logger.info(
                f"""Git Repo to commit the configuration code is not provided. The configuration code can be found at
                {k8s_configuration_path}. You can save the code or commit it to Git for future reference."""
            )
            return

        # Commit the configuration code to Git
        target_repo = GitRepo(
            repo_url=response["target_repo"]["url"],
            git_ref=response["target_repo"]["branch"],
            username=response["target_repo"]["username"],
            password=response["target_repo"]["password"],
        )
        target_repo_path = response["target_repo"]["k8s_path"]
        is_code_committed = self.__commit_k8s_configuration_code_to_git(
            k8s_configuration_path, target_repo, target_repo_path
        )
        # Configure ArgoCD to track the repo as a git repository if the user committed the code
        # in the previous step
        if is_code_committed:
            self.__add_ubermold_argocd_app(target_repo, target_repo_path)



# https://docs.python.org/3/howto/logging.html#library-config
import logging

from rich.logging import RichHandler

logger = logging.getLogger("servicefoundry")


def add_cli_handler(level: int = logging.INFO, show_path=False):
    # TODO (chiragjn): Probably don't use rich handler, it adds hard breaks
    # See: https://github.com/Textualize/rich/discussions/344
    # Maybe try: https://github.com/pycontribs/enrich/blob/main/src/enrich/logging.py
    # or simpler override of logging.Handler with console.print(rich.Text("..."))
    handler = RichHandler(level=level, show_path=show_path)
    handler.setLevel(level)
    logger.addHandler(handler)
    logger.setLevel(logging.DEBUG)



import argparse
import time


def print_numbers(upto: int):
    for i in range(1, upto + 1):
        print(i)
        time.sleep(1)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--upto", type=int)
    args = parser.parse_args()

    print_numbers(args.upto)



import json

import click

from servicefoundry.function_service.app import build_and_run_app
from servicefoundry.function_service.route import RouteGroups


####
# I am keeping the CLI here for now. This will get refactored going forward.
@click.group()
def cli():
    pass


# python -m servicefoundry.function_service run --port 8000 --route-group-json ""
# I am going to use this as "CMD" in the dockerfile to bring up the service
@cli.command()
@click.option("--route-groups-json", required=True, type=str)
@click.option("--port", required=True, type=int)
def run(port: int, route_groups_json: str):
    route_groups = RouteGroups.parse_obj(json.loads(route_groups_json))
    build_and_run_app(route_groups=route_groups, port=port)


if __name__ == "__main__":
    cli()



import functools
import sys
import zipfile

import rich_click as click
from packaging.version import parse as parse_version
from requests.exceptions import ConnectionError
from rich.padding import Padding
from rich.panel import Panel

from servicefoundry.cli.console import console
from servicefoundry.lib.exceptions import BadRequestException, ConfigurationException
from servicefoundry.lib.util import is_debug_env_set


def setup_rich_click():
    click.rich_click.STYLE_ERRORS_SUGGESTION = "blue italic"
    click.rich_click.SHOW_ARGUMENTS = True
    click.rich_click.USE_RICH_MARKUP = True
    click.rich_click.STYLE_HELPTEXT = ""


def handle_exception(exception):
    if is_debug_env_set():
        console.print_exception(show_locals=True)
    if isinstance(exception, BadRequestException):
        print_error(
            f"[cyan bold]status_code[/] {exception.status_code}\n"
            f"[cyan bold]message[/]     {exception.message}"
        )
    elif isinstance(exception, ConnectionError):
        print_error(f"Couldn't connect to Servicefoundry.")
    elif isinstance(exception, ConfigurationException):
        print_error(f"[cyan bold]message[/]     {exception.message}")
    else:
        console.print(f"[red][bold]Error:[/] {str(exception)}[/]")


def handle_exception_wrapper(fn):
    @functools.wraps(fn)
    def inner(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except Exception as e:
            handle_exception(e)
            sys.exit(1)

    return inner


def print_error(message):
    text = Padding(message, (0, 1))
    console.print(
        Panel(
            text,
            border_style="red",
            title="Command failed",
            title_align="left",
            width=click.rich_click.MAX_WIDTH,
        )
    )


def print_message(message):
    text = Padding(message, (0, 1))
    console.print(
        Panel(
            text,
            border_style="cyan",
            title="Success",
            title_align="left",
            width=click.rich_click.MAX_WIDTH,
        )
    )


def unzip_package(path_to_package, destination):
    with zipfile.ZipFile(path_to_package, "r") as zip_ref:
        zip_ref.extractall(destination)


def _prompt_if_no_value_and_supported(prompt: str, hide_input: bool = True):
    import click as _click

    kwargs = {}
    if parse_version(_click.__version__).major >= 8:
        kwargs = dict(prompt=prompt, hide_input=hide_input, prompt_required=False)

    return kwargs



from IPython.display import display
from ipywidgets import Output, widgets
from pygments import highlight
from pygments.formatters import TerminalFormatter
from pygments.lexers import DiffLexer, PythonLexer, TextLexer

from servicefoundry.io.output_callback import OutputCallBack


class NotebookOutputCallBack(OutputCallBack):
    def __init__(self):
        self.output: Output = None
        self.formatter = TerminalFormatter()

    def print_header(self, line):
        line = f"[{line}]"
        if self.output:
            with self.output:
                print(line)
        else:
            print(line)

    def print_line(self, line, lexer=None):
        if self.output:
            if lexer:
                line = highlight(line, lexer, self.formatter)
            with self.output:
                print(line)
        else:
            print(line)

    def print_lines_in_panel(self, lines, header=None):
        self.start_panel()
        self.print_header(header)
        for line in lines:
            self.print_line(line)
        self.close_panel()

    def print_code_in_panel(self, lines, header=None, lang="python"):
        if lang == "python":
            lexer = PythonLexer()
        elif lang == "diff":
            lexer = DiffLexer()
        else:
            lexer = TextLexer()
        self.start_panel()
        if header:
            self.print_header(header)
        for line in lines:
            self.print_line(line, lexer=lexer)
        self.close_panel()

    def start_panel(self):
        output = Output(
            layout=widgets.Layout(width="100%", height="auto", border="1px solid black")
        )
        box = widgets.Box(
            children=[output], layout=widgets.Layout(width="100%", height="auto")
        )
        display(box)
        self.output = output

    def close_panel(self):
        self.output = None



import os
import platform
import shutil
import stat
import tarfile
import urllib.request
import zipfile
from enum import Enum

from servicefoundry.lib.const import (
    HELM_VERSION,
    KUBECTL_VERSION,
    NSC_VERSION,
    TERRAFORM_VERSION,
    TERRAGRUNT_VERSION,
)


class BinaryName(Enum):
    TERRAFORM = "terraform"
    TERRAGRUNT = "terragrunt"
    HELM = "helm"
    KUBECTL = "kubectl"
    NSC = "nsc"


class OsType(Enum):
    WINDOWS = "windows"
    LINUX = "linux"
    DARWIN = "darwin"


class BinaryDependencies:
    def __init__(self):
        # folder containing all the binaries
        self.dir = os.path.join(os.path.expanduser("~"), ".truefoundry", "bin")
        if not os.path.exists(self.dir):
            os.makedirs(self.dir)
        self.ostype = platform.system().lower()
        self.processor = platform.machine().lower()
        self.urls_map = {
            BinaryName.TERRAFORM: {
                OsType.WINDOWS.value: f"https://releases.hashicorp.com/terraform/{TERRAFORM_VERSION}/terraform_{TERRAFORM_VERSION}_windows_{self.processor}.zip",
                OsType.LINUX.value: f"https://releases.hashicorp.com/terraform/{TERRAFORM_VERSION}/terraform_{TERRAFORM_VERSION}_linux_{self.processor}.zip",
                OsType.DARWIN.value: f"https://releases.hashicorp.com/terraform/{TERRAFORM_VERSION}/terraform_{TERRAFORM_VERSION}_darwin_{self.processor}.zip",
            },
            BinaryName.TERRAGRUNT: {
                OsType.WINDOWS.value: f"https://github.com/gruntwork-io/terragrunt/releases/download/v{TERRAGRUNT_VERSION}/terragrunt_windows_{self.processor}.exe",
                OsType.LINUX.value: f"https://github.com/gruntwork-io/terragrunt/releases/download/v{TERRAGRUNT_VERSION}/terragrunt_linux_{self.processor}",
                OsType.DARWIN.value: f"https://github.com/gruntwork-io/terragrunt/releases/download/v{TERRAGRUNT_VERSION}/terragrunt_darwin_{self.processor}",
            },
            BinaryName.HELM: {
                OsType.WINDOWS.value: f"https://get.helm.sh/helm-v{HELM_VERSION}-windows-{self.processor}.tar.gz",
                OsType.LINUX.value: f"https://get.helm.sh/helm-v{HELM_VERSION}-linux-{self.processor}.tar.gz",
                OsType.DARWIN.value: f"https://get.helm.sh/helm-v{HELM_VERSION}-darwin-{self.processor}.tar.gz",
            },
            BinaryName.KUBECTL: {
                OsType.WINDOWS.value: f"https://dl.k8s.io/release/v{KUBECTL_VERSION}/bin/windows/{self.processor}/kubectl.exe",
                OsType.LINUX.value: f"https://dl.k8s.io/release/v{KUBECTL_VERSION}/bin/linux/{self.processor}/kubectl",
                OsType.DARWIN.value: f"https://dl.k8s.io/release/v{KUBECTL_VERSION}/bin/darwin/{self.processor}/kubectl",
            },
            BinaryName.NSC: {
                OsType.WINDOWS.value: f"https://github.com/nats-io/nsc/releases/download/v{NSC_VERSION}/nsc-windows-{self.processor}.zip",
                OsType.LINUX.value: f"https://github.com/nats-io/nsc/releases/download/v{NSC_VERSION}/nsc-linux-{self.processor}.zip",
                OsType.DARWIN.value: f"https://github.com/nats-io/nsc/releases/download/v{NSC_VERSION}/nsc-darwin-{self.processor}.zip",
            },
        }

    def which(self, binary: BinaryName = None):
        if binary.value == BinaryName.TERRAFORM.value:
            return self.__get_terraform_path()
        elif binary.value == BinaryName.TERRAGRUNT.value:
            return self.__get_terragrunt_path()
        elif binary.value == BinaryName.HELM.value:
            return self.__get_helm_path()
        elif binary.value == BinaryName.KUBECTL.value:
            return self.__get_kubectl_path()
        elif binary.value == BinaryName.NSC.value:
            return self.__get_nsc_path()
        else:
            raise Exception(
                f"{binary.value} not present. Please install it first and then run the script."
            )

    # Funtion to add executable permission to the binary
    def __add_executable_permission(self, path_to_executable):
        state = os.stat(path_to_executable)
        os.chmod(path_to_executable, state.st_mode | stat.S_IEXEC)

    def __get_terraform_path(self):
        terraform_path = os.path.join(self.dir, BinaryName.TERRAFORM.value)
        # check if terraform already exists
        if os.path.exists(terraform_path):
            return terraform_path

        # get url to download binary, based on os type and processor type
        url = self.urls_map[BinaryName.TERRAFORM][self.ostype]
        file_name = os.path.join(self.dir, "terraform.zip")
        urllib.request.urlretrieve(url, file_name)

        # extract the dowloaded zip and move the binary to destination
        with zipfile.ZipFile(file_name, "r") as zip_file:
            zip_file.extractall(self.dir)
            # clean unwanted files
            os.remove(file_name)
        self.__add_executable_permission(terraform_path)

        return terraform_path

    def __get_terragrunt_path(self):
        terragrunt_path = os.path.join(self.dir, BinaryName.TERRAGRUNT.value)
        if os.path.exists(terragrunt_path):
            return terragrunt_path

        url = self.urls_map[BinaryName.TERRAGRUNT][self.ostype]
        urllib.request.urlretrieve(
            url, os.path.join(self.dir, BinaryName.TERRAGRUNT.value)
        )

        self.__add_executable_permission(terragrunt_path)
        return terragrunt_path

    def __get_helm_path(self):
        #  return local installation if present
        helm_path = shutil.which("helm")
        if helm_path:
            return helm_path
        helm_path = os.path.join(self.dir, BinaryName.HELM.value)
        if os.path.exists(helm_path):
            return helm_path

        url = self.urls_map[BinaryName.HELM][self.ostype]
        file_name = os.path.join(self.dir, "helm.tar.gz")
        urllib.request.urlretrieve(url, file_name)

        with tarfile.open(file_name, "r:gz") as tar_file:
            dir_name = os.path.join(self.dir, tar_file.getnames()[0])
            tar_file.extractall(self.dir)
            shutil.move(
                os.path.join(dir_name, BinaryName.HELM.value),
                os.path.join(self.dir, BinaryName.HELM.value),
            )
            shutil.rmtree(dir_name)
            os.remove(file_name)
        self.__add_executable_permission(helm_path)
        return helm_path

    def __get_kubectl_path(self):
        kubectl_path = shutil.which("kubectl")
        if kubectl_path:
            return kubectl_path
        kubectl_path = os.path.join(self.dir, BinaryName.KUBECTL.value)
        if os.path.exists(kubectl_path):
            return kubectl_path

        url = self.urls_map[BinaryName.KUBECTL][self.ostype]
        urllib.request.urlretrieve(
            url, os.path.join(self.dir, BinaryName.KUBECTL.value)
        )

        self.__add_executable_permission(kubectl_path)
        return kubectl_path

    def __get_nsc_path(self):
        nsc_path = os.path.join(self.dir, BinaryName.NSC.value)
        if os.path.exists(nsc_path):
            return nsc_path

        url = self.urls_map[BinaryName.NSC][self.ostype]
        file_name = os.path.join(self.dir, "nsc.zip")
        urllib.request.urlretrieve(url, file_name)

        with zipfile.ZipFile(file_name, "r") as zip_file:
            zip_file.extractall(self.dir)
            os.remove(file_name)
        self.__add_executable_permission(nsc_path)

        return nsc_path



from servicefoundry.core.login import login
from servicefoundry.core.logout import logout
from servicefoundry.io.notebook.notebook_util import is_notebook

if is_notebook():
    try:
        import ipywidgets
    except ImportError:
        print("Run `pip install ipywidgets` to use notebook features.")

__all__ = [
    "login",
    "logout",
]



import os
from tempfile import TemporaryDirectory
from typing import Optional

from servicefoundry.auto_gen.models import DockerFileBuild, PythonBuild
from servicefoundry.builder.builders import dockerfile
from servicefoundry.builder.builders.tfy_python_buildpack.dockerfile_template import (
    generate_dockerfile_content,
)

__all__ = ["generate_dockerfile_content", "build"]


def _convert_to_dockerfile_build_config(
    build_configuration: PythonBuild,
    dockerfile_path: str,
) -> DockerFileBuild:
    dockerfile_content = generate_dockerfile_content(
        build_configuration=build_configuration
    )
    with open(dockerfile_path, "w", encoding="utf8") as fp:
        fp.write(dockerfile_content)

    return DockerFileBuild(
        type="dockerfile",
        dockerfile_path=dockerfile_path,
        build_context_path=build_configuration.build_context_path,
    )


def build(
    tag: str,
    build_configuration: PythonBuild,
    cache_from: Optional[str] = None,
):
    with TemporaryDirectory() as local_dir:
        docker_build_configuration = _convert_to_dockerfile_build_config(
            build_configuration, dockerfile_path=os.path.join(local_dir, "Dockerfile")
        )
        dockerfile.build(
            tag=tag,
            build_configuration=docker_build_configuration,
            cache_from=cache_from,
        )



from typing import Callable, Dict

from servicefoundry.builder.builders import dockerfile, tfy_python_buildpack

BUILD_REGISTRY: Dict[str, Callable] = {
    "dockerfile": dockerfile.build,
    "tfy-python-buildpack": tfy_python_buildpack.build,
}

__all__ = ["get_builder"]


def get_builder(build_configuration_type: str) -> Callable:
    if build_configuration_type not in BUILD_REGISTRY:
        raise NotImplementedError(f"Builder for {build_configuration_type} not found")

    return BUILD_REGISTRY[build_configuration_type]



import enum
import re
from typing import Union

from pydantic import Field, constr, root_validator
from typing_extensions import Literal

from servicefoundry.auto_gen import models

# TODO (chiragjn): Setup a base class for auto_gen.models to make `extra = "forbid"` default


class CUDAVersion(str, enum.Enum):
    CUDA_9_2_CUDNN7 = "9.2-cudnn7"
    CUDA_10_0_CUDNN7 = "10.0-cudnn7"
    CUDA_10_1_CUDNN7 = "10.1-cudnn7"
    CUDA_10_1_CUDNN8 = "10.1-cudnn8"
    CUDA_10_2_CUDNN7 = "10.2-cudnn7"
    CUDA_10_2_CUDNN8 = "10.2-cudnn8"
    CUDA_11_0_CUDNN8 = "11.0-cudnn8"
    CUDA_11_1_CUDNN8 = "11.1-cudnn8"
    CUDA_11_2_CUDNN8 = "11.2-cudnn8"
    CUDA_11_3_CUDNN8 = "11.3-cudnn8"
    CUDA_11_4_CUDNN8 = "11.4-cudnn8"
    CUDA_11_5_CUDNN8 = "11.5-cudnn8"
    CUDA_11_6_CUDNN8 = "11.6-cudnn8"
    CUDA_11_7_CUDNN8 = "11.7-cudnn8"
    CUDA_11_8_CUDNN8 = "11.8-cudnn8"


class GPUType(str, enum.Enum):
    K80 = "K80"
    V100 = "V100"
    T4 = "T4"
    A10G = "A10G"
    A100_40GB = "A100_40GB"
    A100_80GB = "A100_80GB"


class DockerFileBuild(models.DockerFileBuild):
    class Config:
        extra = "forbid"

    type: constr(regex=r"dockerfile") = "dockerfile"


class PythonBuild(models.PythonBuild):
    class Config:
        extra = "forbid"

    type: constr(regex=r"tfy-python-buildpack") = "tfy-python-buildpack"

    @root_validator
    def validate_python_version_when_cuda_version(cls, values):
        if values.get("cuda_version"):
            python_version = values.get("python_version")
            if python_version and not re.match(r"^3\.\d+$", python_version):
                raise ValueError(
                    f'`python_version` must be 3.x (e.g. "3.9") when `cuda_version` field is '
                    f"provided but got {python_version!r}. If you are adding a "
                    f'patch version, please remove it (e.g. "3.9.2" should be "3.9")'
                )
        return values


class RemoteSource(models.RemoteSource):
    class Config:
        extra = "forbid"

    type: constr(regex=r"remote") = "remote"


class LocalSource(models.LocalSource):
    class Config:
        extra = "forbid"

    type: constr(regex=r"local") = "local"


class Build(models.Build):
    class Config:
        extra = "forbid"

    type: constr(regex=r"build") = "build"
    build_source: Union[
        models.RemoteSource, models.GitSource, models.LocalSource
    ] = Field(default_factory=LocalSource)


class Manual(models.Manual):
    class Config:
        extra = "forbid"

    type: constr(regex=r"manual") = "manual"


class Schedule(models.Schedule):
    class Config:
        extra = "forbid"

    type: constr(regex=r"scheduled") = "scheduled"


class GitSource(models.GitSource):
    class Config:
        extra = "forbid"

    type: constr(regex=r"git") = "git"


class HttpProbe(models.HttpProbe):
    class Config:
        extra = "forbid"

    type: constr(regex=r"http") = "http"


class BasicAuthCreds(models.BasicAuthCreds):
    class Config:
        extra = "forbid"

    type: constr(regex=r"basic_auth") = "basic_auth"


class TruefoundryModelRegistry(models.TruefoundryModelRegistry):
    class Config:
        extra = "forbid"

    type: constr(regex=r"tfy-model-registry") = "tfy-model-registry"


class HuggingfaceModelHub(models.HuggingfaceModelHub):
    class Config:
        extra = "forbid"

    type: constr(regex=r"hf-model-hub") = "hf-model-hub"


class HealthProbe(models.HealthProbe):
    class Config:
        extra = "forbid"


class Image(models.Image):
    class Config:
        extra = "forbid"

    type: constr(regex=r"image") = "image"


class Port(models.Port):
    class Config:
        extra = "forbid"


class Resources(models.Resources):
    class Config:
        extra = "forbid"


class Param(models.Param):
    class Config:
        extra = "forbid"


class FileMount(models.FileMount):
    class Config:
        extra = "forbid"


class CPUUtilizationMetric(models.CPUUtilizationMetric):
    class Config:
        extra = "forbid"

    type: Literal["cpu_utilization"] = "cpu_utilization"


class RPSMetric(models.RPSMetric):
    class Config:
        extra = "forbid"

    type: Literal["rps"] = "rps"


class Autoscaling(models.Autoscaling):
    class Config:
        extra = "forbid"


class GPU(models.GPU):
    class Config:
        extra = "forbid"

    type: Union[GPUType, str] = Field(
        ...,
        description="+label GPU Name\n+usage=Name of the Nvidia GPU Type. One of `servicefoundry.GPUType`",
    )


class BlueGreen(models.BlueGreen):
    class Config:
        extra = "forbid"

    type: Literal["blue_green"] = Field("blue_green", description="+value=blue_green")


class Canary(models.Canary):
    class Config:
        extra = "forbid"

    type: Literal["canary"] = Field("canary", description="+value=canary")


class Rolling(models.Rolling):
    class Config:
        extra = "forbid"

    type: Literal["rolling_update"] = Field(
        "rolling_update", description="+value=rolling_update"
    )



import rich_click as click

from servicefoundry.cli.console import console
from servicefoundry.cli.const import COMMAND_CLS
from servicefoundry.cli.util import (
    _prompt_if_no_value_and_supported,
    handle_exception_wrapper,
)
from servicefoundry.io.rich_output_callback import RichOutputCallBack
from servicefoundry.lib.messages import PROMPT_POST_LOGIN
from servicefoundry.lib.session import login


@click.command(name="login", cls=COMMAND_CLS)
@click.option("--relogin", type=click.BOOL, is_flag=True, default=False)
@click.option("--host", type=click.STRING, default=None)
@click.option(
    "--api-key",
    "--api_key",
    type=click.STRING,
    default=None,
    **_prompt_if_no_value_and_supported(prompt="API Key", hide_input=True),
)
@handle_exception_wrapper
def login_command(relogin: bool, host: str, api_key: str):
    """
    Login to Truefoundry
    """
    callback = RichOutputCallBack()
    login(api_key=api_key, host=host, relogin=relogin, output_hook=callback)
    console.print(PROMPT_POST_LOGIN)


def get_login_command():
    return login_command



from typing import Dict, Optional, Union

from pydantic import BaseModel

from servicefoundry.auto_gen.models import DockerFileBuild, PythonBuild
from servicefoundry.builder.builders import get_builder


class _BuildConfig(BaseModel):
    # I cannot use Field(discriminator="build_config_type") here as
    # build_config_type in the build configs is not a Literal.
    __root__: Union[DockerFileBuild, PythonBuild]


def build(
    build_configuration: Union[BaseModel, Dict],
    tag: str,
    cache_from: Optional[str] = None,
):
    build_configuration = _BuildConfig.parse_obj(build_configuration).__root__
    builder = get_builder(build_configuration.type)
    return builder(
        build_configuration=build_configuration,
        tag=tag,
        cache_from=cache_from,
    )


if __name__ == "__main__":
    # TODO: remove these and write tests for Build class to dockerfile content.
    import os
    from tempfile import TemporaryDirectory

    with TemporaryDirectory() as local_dir:
        dockerfile_path = os.path.join(local_dir, "Dockerfile.test")
        with open(dockerfile_path, "w", encoding="utf8") as fp:
            fp.write("from postgres:latest")

        build_config = DockerFileBuild(
            type="dockerfile",
            build_context_path=local_dir,
            dockerfile_path=dockerfile_path,
        )

        build(tag="docker-test", build_configuration=build_config)

    with TemporaryDirectory() as local_dir:
        requirements_path = os.path.join(local_dir, "requirements.txt")
        with open(requirements_path, "w", encoding="utf8") as fp:
            fp.write("requests")

        python_path = os.path.join(local_dir, "main.py")
        with open(python_path, "w", encoding="utf8") as fp:
            fp.write("import requests; print(requests.__version__);")

        build_config = PythonBuild(
            type="tfy-python-buildpack",
            build_context_path=local_dir,
            command=["python", "main.py"],
            python_version="3.7.13",
        )
        build(tag="python-test-exec", build_configuration=build_config)
        build_config = PythonBuild(
            type="tfy-python-buildpack",
            build_context_path=local_dir,
            command="python main.py",
            python_version="3.8",
        )
        build(tag="python-test-shell", build_configuration=build_config)

    with TemporaryDirectory() as local_dir:
        requirements_path = os.path.join(local_dir, "requirements.txt")
        with open(requirements_path, "w", encoding="utf8") as fp:
            fp.write("numpy")

        python_path = os.path.join(local_dir, "main.py")
        with open(python_path, "w", encoding="utf8") as fp:
            fp.write("import numpy; import requests; print(requests.__version__);")

        build_config = PythonBuild(
            type="tfy-python-buildpack",
            build_context_path=local_dir,
            command=["python", "main.py"],
            python_version="3.7.13",
            pip_packages=["requests"],
        )
        build(tag="python-pip-and-req-file", build_configuration=build_config)

    with TemporaryDirectory() as local_dir:
        python_path = os.path.join(local_dir, "main.py")
        with open(python_path, "w", encoding="utf8") as fp:
            fp.write("import requests; print(requests.__version__);")

        build_config = PythonBuild(
            type="tfy-python-buildpack",
            build_context_path=local_dir,
            command=["python", "main.py"],
            python_version="3.7.13",
            pip_packages=["requests>0.1"],
        )
        build(tag="python-only-pip", build_configuration=build_config)

    with TemporaryDirectory() as local_dir:
        python_path = os.path.join(local_dir, "main.py")
        with open(python_path, "w", encoding="utf8") as fp:
            fp.write("print(1)")

        build_config = PythonBuild(
            type="tfy-python-buildpack",
            build_context_path=local_dir,
            command=["python", "main.py"],
            python_version="3.7.13",
        )
        build(tag="python-only-no-pip-no-req", build_configuration=build_config)



from threading import Thread
from typing import Any, Callable, Dict, Optional

import yaml

from servicefoundry.auto_gen.models import Port, Resources
from servicefoundry.function_service.app import build_and_run_app_in_background_thread
from servicefoundry.function_service.build import BuildConfig
from servicefoundry.function_service.remote import RemoteClass
from servicefoundry.function_service.route import RouteGroups
from servicefoundry.v2.lib.deployable_patched_models import Service
from servicefoundry.v2.lib.patched_models import Build, LocalSource


class FunctionService:
    def __init__(
        self,
        name: str,
        build_config: Optional[BuildConfig] = None,
        resources: Optional[Resources] = None,
        replicas: int = 1,
        port: int = 8000,
        env: Optional[Dict[str, str]] = None,
    ):
        self._name = name
        self._build_config = build_config or BuildConfig()
        self._resources = resources or Resources()
        self._replicas = replicas
        self._port = port
        self._env = env or {}

        self._route_groups: RouteGroups = RouteGroups()

    @property
    def route_groups(self) -> RouteGroups:
        return self._route_groups

    def __repr__(self):
        return yaml.dump(
            dict(
                name=self._name,
                build_config=self._build_config.dict(),
                resources=self._resources.dict(),
                routes=self._route_groups.dict(),
                replicas=self._replicas,
                port=self._port,
                env=self._env,
            ),
            indent=2,
        )

    def register_function(
        self,
        func: Callable,
        *,
        path: Optional[str] = None,
    ):
        self._route_groups.register_function(func=func, path=path)

    def register_class(
        self,
        class_,
        *,
        init_kwargs: Optional[Dict[str, Any]] = None,
        name: Optional[str] = None,
    ):
        # TODO: I need to rethink this `RemoteClass`.
        #   I am mixing up multiple responsibilities here.
        #   For now, I am removing the burden of using `remote` from the user when deploying
        #   an instance of a class.
        remote_class = RemoteClass(class_, init_kwargs=init_kwargs, name=name)
        self._route_groups.register_class(remote_class=remote_class)

    def run(self) -> Thread:
        return build_and_run_app_in_background_thread(
            route_groups=self._route_groups, port=self._port
        )

    def get_deployment_definition(self) -> Service:
        # Keeping this function right now so that later,
        # the constructor of the application call this function
        # to get the component spec, if an object of this class
        # is directly passed as an component
        tfy_python_build_config = self._build_config.to_tfy_python_build_config(
            port=self._port, route_groups=self._route_groups
        )
        service = Service(
            name=self._name,
            image=Build(build_source=LocalSource(), build_spec=tfy_python_build_config),
            resources=self._resources,
            replicas=self._replicas,
            ports=[Port(port=self._port, expose=True)],
            env=self._env,
        )
        return service

    def deploy(self, workspace_fqn: str, wait: bool = True):
        service = self.get_deployment_definition()
        service.deploy(workspace_fqn=workspace_fqn, wait=wait)



import logging
from typing import Optional

import rich_click as click

from servicefoundry.cli.const import COMMAND_CLS, GROUP_CLS
from servicefoundry.cli.display_util import print_obj
from servicefoundry.cli.util import handle_exception_wrapper
from servicefoundry.lib.dao import workspace as workspace_lib
from servicefoundry.logger import logger


@click.group(name="create", cls=GROUP_CLS)
def create_command():
    # TODO (chiragjn): Figure out a way to update supported resources based on ENABLE_* flags
    """
    Create Truefoundry resources

    \b
    Supported resources:
    - Workspace
    """
    pass


@click.command(name="workspace", cls=COMMAND_CLS, help="Create a new Workspace")
@click.argument("name", type=click.STRING)
@click.option(
    "-c",
    "--cluster-name",
    "--cluster_name",
    type=click.STRING,
    required=True,
    help="Cluster to create this workspace in",
)
@click.option(
    "--cpu-limit",
    "--cpu_limit",
    type=click.FLOAT,
    default=None,
    help="CPU Limit",
)
@click.option(
    "--memory-limit",
    "--memory_limit",
    type=click.INT,
    default=None,
    help="Memory Limit in MB",
)
@click.option(
    "--ephemeral-storage-limit",
    "--ephemeral_storage_limit",
    type=click.INT,
    default=None,
    help="Ephemeral Storage Limit in GB",
)
@handle_exception_wrapper
def create_workspace(
    name: str,
    cluster_name: str,
    cpu_limit: Optional[float],
    memory_limit: Optional[int],
    ephemeral_storage_limit: Optional[int],
):
    workspace = workspace_lib.create_workspace(
        name=name,
        cpu_limit=cpu_limit,
        memory_limit=memory_limit,
        ephemeral_storage_limit=ephemeral_storage_limit,
        cluster_name=cluster_name,
    )
    print_obj("Workspace", workspace.dict())


def get_create_command():
    create_command.add_command(create_workspace)
    return create_command



import os

import questionary
from questionary import Choice

from servicefoundry.io.input_hook import InputHook
from servicefoundry.io.parameters import OptionsParameter, Parameter

MSG_CREATE_NEW_SPACE = "Create a new workspace"


def python_file_filter(path: str):
    if path.endswith(".py"):
        return True
    return False


# TODO (chiragjn): This needs to support validations


class CliInputHook(InputHook):
    def __init__(self):
        super().__init__()

    def _get_default(self, param):
        if param.default:
            return str(param.default)
        if param.suggest:
            return str(param.suggest)
        return ""

    def confirm(self, prompt: str, default=False):
        return questionary.confirm(prompt, default=default).ask()

    def ask_string(self, param: Parameter):
        return questionary.text(param.prompt, default=self._get_default(param)).ask()

    def ask_number(self, param: Parameter):
        while True:
            value = questionary.text(
                param.prompt, default=self._get_default(param)
            ).ask()
            if value.isdigit():
                return int(value)
            else:
                print(f"Not an integer value {value}. Try again")

    def ask_float(self, param: Parameter):
        while True:
            value = questionary.text(
                param.prompt, default=self._get_default(param)
            ).ask()
            try:
                return float(value)
            except ValueError:
                print(f"Not an float value {value}. Try again")

    def ask_file_path(self, param: Parameter):
        return questionary.path(param.prompt, default=self._get_default(param)).ask()

    def ask_option(self, param: OptionsParameter):
        choices = [
            Choice(title=value, value=key) for key, value in param.options.items()
        ]
        return questionary.select(
            param.prompt, choices=choices, default=param.default
        ).ask()

    def ask_python_file(self, param: Parameter):
        while True:
            path = questionary.path(param.prompt, file_filter=python_file_filter).ask()
            if os.path.isfile(path):
                return path
            else:
                print(f"Incorrect path {path}. Try again.")



import time

import requests

from servicefoundry.lib.clients.utils import poll_for_function, request_handling
from servicefoundry.lib.const import VERSION_PREFIX
from servicefoundry.lib.model.entity import DeviceCode, Token
from servicefoundry.logger import logger


class AuthServiceClient:
    def __init__(self, base_url):
        from servicefoundry.lib.clients.service_foundry_client import (
            ServiceFoundryServiceClient,
        )

        client = ServiceFoundryServiceClient(init_session=False, base_url=base_url)
        tenant_info = client.get_tenant_info()

        self._auth_server_url = tenant_info.auth_server_url
        self._tenant_name = tenant_info.tenant_name

    def refresh_token(self, token: Token) -> Token:
        if not token.refresh_token:
            # TODO: Add a way to propagate error messages without traceback to the output interface side
            raise Exception(
                "Unable to resume login session. Please log in again using `sfy login [--host HOST] --relogin`"
            )
        url = f"{self._auth_server_url}/api/{VERSION_PREFIX}/oauth/token/refresh"
        data = {
            "tenantName": token.tenant_name,
            "refreshToken": token.refresh_token,
        }
        res = requests.post(url, data=data)
        res = request_handling(res)
        return Token.parse_obj(res)

    def get_device_code(self) -> DeviceCode:
        url = f"{self._auth_server_url}/api/{VERSION_PREFIX}/oauth/device"
        data = {"tenantName": self._tenant_name}
        res = requests.post(url, data=data)
        res = request_handling(res)
        return DeviceCode.parse_obj(res)

    def get_token_from_device_code(
        self, device_code: str, timeout: float = 60
    ) -> Token:
        url = f"{self._auth_server_url}/api/{VERSION_PREFIX}/oauth/device/token"
        data = {
            "tenantName": self._tenant_name,
            "deviceCode": device_code,
        }
        start_time = time.monotonic()
        poll_interval_seconds = 1

        for response in poll_for_function(
            requests.post, poll_after_secs=poll_interval_seconds, url=url, data=data
        ):
            if response.status_code == 201:
                response = response.json()
                return Token.parse_obj(response)
            elif response.status_code == 202:
                logger.debug("User has not authorized yet. Checking again.")
            else:
                raise Exception(
                    "Failed to get token using device code. "
                    f"status_code {response.status_code},\n {response.text}"
                )
            time_elapsed = time.monotonic() - start_time
            if time_elapsed > timeout:
                logger.warning("Polled server for %s secs.", int(time_elapsed))
                break

        raise Exception(f"Did not get authorized within {timeout} seconds.")



import argparse
import logging

from servicefoundry import ModelDeployment, Resources, TruefoundryModelRegistry

logging.basicConfig(level=logging.INFO)

parser = argparse.ArgumentParser()
parser.add_argument("--workspace-fqn", type=str, required=True)
parser.add_argument("--model-fqn", type=str, required=True)
args = parser.parse_args()

service = ModelDeployment(
    name="from-cli-mlf",
    model_source=TruefoundryModelRegistry(model_version_fqn=args.model_fqn),
    resources=Resources(cpu_limit=1.0, memory_limit=600),
)
service.deploy(workspace_fqn=args.workspace_fqn)



import argparse
import logging

from servicefoundry import HuggingfaceModelHub, ModelDeployment, Resources

logging.basicConfig(level=logging.INFO)

parser = argparse.ArgumentParser()
parser.add_argument("--workspace-fqn", type=str, required=True)
parser.add_argument("--model-id", type=str, required=True)
parser.add_argument("--pipeline", type=str, default=None)
args = parser.parse_args()

service = ModelDeployment(
    name="from-cli-hf",
    model_source=HuggingfaceModelHub(model_id=args.model_id, pipeline=args.pipeline),
    resources=Resources(cpu_limit=1.0, memory_limit=600),
)
service.deploy(workspace_fqn=args.workspace_fqn)



import numpy as np

from servicefoundry.function_service import remote


def preprocess(image_url: str):
    return np.random.uniform(0, 1, [1, 1, 1]).tolist()


class Model:
    def __init__(self, path: str):
        self.path = path

    def predict(self, image_url: str):
        # preprocessed_data = preprocess(image_url=image_url)

        preprocessed_data = remote(preprocess)(image_url=image_url)

        print(f"running prediction on {image_url} using model from {self.path}")
        return {"class": "1"}



import json
import logging

import rich_click as click

from servicefoundry import builder
from servicefoundry.cli.const import GROUP_CLS
from servicefoundry.cli.util import handle_exception_wrapper
from servicefoundry.logger import logger


@click.group(
    name="build",
    cls=GROUP_CLS,
    invoke_without_command=True,
    help="Build docker image locally from Truefoundry spec",
)
@click.option(
    "--name",
    type=click.STRING,
    required=True,
    help="Name for the image being build - used as docker tag",
)
@click.option(
    "--build-config",
    "--build_config",
    type=click.STRING,
    required=True,
    help="Build part of the spec as a json spec",
)
@click.option(
    "--cache",
    type=click.STRING,
    default=None,
    help="Docker repository to use as a cache to pull layers from",
)
@handle_exception_wrapper
def build_command(name, cache, build_config):
    if build_config:
        builder.build(
            build_configuration=json.loads(build_config), tag=name, cache_from=cache
        )
        return


def get_build_command():
    return build_command



import datetime
import functools

from rich import box
from rich import print_json as _rich_print_json
from rich.table import Table

from servicefoundry.cli.config import CliConfig
from servicefoundry.cli.console import console
from servicefoundry.cli.const import DISPLAY_DATETIME_FORMAT
from servicefoundry.json_util import json_default_encoder

print_json = functools.partial(
    _rich_print_json, highlight=False, default=json_default_encoder
)


def get_table(title):
    return Table(title=title, show_lines=False, safe_box=True, box=box.MINIMAL)


def stringify(x):
    if isinstance(x, datetime.datetime):
        return x.astimezone().strftime(DISPLAY_DATETIME_FORMAT)
    elif isinstance(x, str):
        return x
    else:
        return str(x)


def print_list(title, items, columns=None):
    if CliConfig.get("json"):
        print_json(data=items)
        return

    table = get_table(title)

    if items:
        if not columns:
            columns = items[0].keys()
        for column in columns:
            no_wrap = column in ["id", "fqn"]
            table.add_column(column, justify="left", overflow="fold", no_wrap=no_wrap)

    for item in items:
        row = []
        for c in columns:
            row.append(stringify(item[c]))
        table.add_row(*row)
    console.print(table)


def print_entity_list(title, items):
    if CliConfig.get("json"):
        print_json(data=items)
        return

    table = get_table(title)

    columns = []
    if items:
        columns = items[0].list_row_data().keys()
        for column in columns:
            no_wrap = "FQN" in column or "Name" in column
            table.add_column(column, justify="left", overflow="fold", no_wrap=no_wrap)

    for item in items:
        entity = item.list_row_data()
        row = []
        for c in columns:
            row.append(stringify(entity[c]))
        table.add_row(*row)
    console.print(table)


def print_obj(title, item, columns=None):
    if CliConfig.get("json"):
        print_json(data=item)
        return

    table = get_table(title)

    if not columns:
        columns = item.keys()

    # transpose
    keys, columns = columns, ["key", "value"]

    for column in columns:
        no_wrap = column in ["id", "fqn"]
        table.add_column(column, justify="left", overflow="fold", no_wrap=no_wrap)
    for key in keys:
        table.add_row(f"[bold]{stringify(key)}[/]", stringify(item[key]))
    console.print(table)


def print_entity_obj(title, entity):
    if CliConfig.get("json"):
        print_json(data=entity)
        return

    table = get_table(title)

    columns = entity.get_data().keys()

    # transpose
    keys, columns = columns, ["key", "value"]

    for column in columns:
        no_wrap = "FQN" in column or "Name" in column
        table.add_column(column, justify="left", overflow="fold", no_wrap=no_wrap)
    entity_data = entity.get_data()
    for key in keys:
        table.add_row(f"[bold]{stringify(key)}[/]", stringify(entity_data[key]))
    console.print(table)



import time

from model import Model
from model_composition import ModelComposition
from utils import preprocess

from servicefoundry.function_service import BuildConfig, FunctionService, remote

MODEL_PATHS = {
    "model_1_path": "foo",
    "model_2_path": "bar",
    "model_3_path": "baz",
}

preprocess_service = FunctionService(
    name="preprocess-service",
    build_config=BuildConfig(pip_packages=["numpy<1.22.0"]),
    port=7000,
)
preprocess_service.register_function(preprocess)

model_service = FunctionService(name="model-service", port=7001)
model_service.register_class(
    Model, init_kwargs={"model_path": MODEL_PATHS["model_1_path"]}, name="model_1"
)
model_service.register_class(
    Model,
    init_kwargs={"model_path": MODEL_PATHS["model_2_path"]},
    name="model_2",
)
model_service.register_class(
    Model,
    init_kwargs={"model_path": MODEL_PATHS["model_2_path"]},
    name="model_3",
)

print(preprocess_service)
print(model_service)

preprocess_service.run()
model_service.run()
time.sleep(5)

composed_model = remote(
    ModelComposition,
    init_kwargs=MODEL_PATHS,
)
print(composed_model.predict(image_url="foo"))
time.sleep(3)


composed_model_service = FunctionService(name="composed-model-service", port=7002)
composed_model_service.register_class(ModelComposition, init_kwargs=MODEL_PATHS)
print(composed_model_service)
composed_model_service.run().join()
time.sleep(1)

# 1. Multiple classes, each class can have its own route prefix.
# 2. module vs filepath. Can I import a path directly as a module. This is relative to source.
# 3. What is the FQN for a normal function and a method. which I can deterministically generate at run time.
# 4. What if someone uses the same class and want to init it with two diff model path.



import logging
from datetime import datetime
from typing import Optional

import rich_click as click
from dateutil.tz import tzlocal

from servicefoundry.cli.const import COMMAND_CLS
from servicefoundry.cli.util import handle_exception_wrapper
from servicefoundry.io.rich_output_callback import RichOutputCallBack
from servicefoundry.lib.clients.service_foundry_client import (
    ServiceFoundryServiceClient,
)
from servicefoundry.lib.logs_utils import get_timestamp_from_timestamp_or_duration
from servicefoundry.lib.util import get_deployment_fqn_from_application_fqn
from servicefoundry.logger import logger


@click.command(name="logs", cls=COMMAND_CLS)
@click.option(
    "--application-fqn",
    type=click.STRING,
    help="FQN of the application",
    required=True,
)
@click.option(
    "--version",
    type=click.INT,
    help="Version number of the application deployment",
    default=None,
)
@click.option(
    "--since",
    type=click.STRING,
    help="Show logs since timestamp (e.g.2013-01-02T13:23:37Z) or relative (e.g. 42m for 42 minutes)",
    default="2h",
)
@click.option(
    "--until",
    type=click.STRING,
    help="Show logs until timestamp (e.g.2013-01-02T13:23:37Z) or relative (e.g. 42m for 42 minutes)",
    default=None,
)
@click.option(
    "-n", "--tail", type=click.INT, help="Number of logs to tail", default=None
)
@click.option("-f", "--follow", help="Follow log output", is_flag=True, default=False)
@handle_exception_wrapper
def logs_command(
    application_fqn: str,
    version: Optional[int],
    since: str,
    until: Optional[str],
    tail: Optional[int],
    follow: bool,
) -> None:
    """
    Get logs for Application
    """
    start_ts = get_timestamp_from_timestamp_or_duration(since)
    end_ts = (
        get_timestamp_from_timestamp_or_duration(until)
        if until
        else datetime.now().timestamp() * 1000
    )

    output_hook = RichOutputCallBack()
    tfs_client = ServiceFoundryServiceClient()

    if version:
        deployment_fqn = get_deployment_fqn_from_application_fqn(
            application_fqn=application_fqn, version=version
        )
        deployment_fqn_response = tfs_client.get_deployment_info_by_fqn(deployment_fqn)
        workspace_id, application_id, deployment_id = (
            deployment_fqn_response.workspaceId,
            deployment_fqn_response.applicationId,
            deployment_fqn_response.deploymentId,
        )
    else:
        app_fqn_response = tfs_client.get_application_info_by_fqn(application_fqn)
        workspace_id, application_id, deployment_id = (
            app_fqn_response.workspaceId,
            app_fqn_response.applicationId,
            None,
        )

    if until:
        logger.info(
            "Fetching logs from %s to %s for application fqn: %s",
            datetime.fromtimestamp(start_ts / 1000, tzlocal()).isoformat(),
            datetime.fromtimestamp(end_ts / 1000, tzlocal()).isoformat(),
            application_fqn,
        )
    else:
        logger.info(
            "Fetching logs from %s for application fqn: %s",
            datetime.fromtimestamp(start_ts / 1000, tzlocal()).isoformat(),
            application_fqn,
        )

    if until or not follow:
        tfs_client.fetch_deployment_logs(
            workspace_id=workspace_id,
            application_id=application_id,
            deployment_id=deployment_id,
            start_ts=start_ts,
            end_ts=end_ts,
            limit=tail,
            callback=output_hook,
        )
    else:
        tfs_client.poll_logs_for_deployment(
            workspace_id=workspace_id,
            application_id=application_id,
            deployment_id=deployment_id,
            start_ts=start_ts,
            limit=tail,
            poll_interval_seconds=5,
            callback=output_hook,
        )


def get_logs_command():
    return logs_command



import json
import os

import rich_click as click

from servicefoundry.cli.const import COMMAND_CLS, GROUP_CLS
from servicefoundry.lib.infra.install_truefoundry import InfraController
from servicefoundry.logger import logger


@click.group(
    name="bootstrap",
    cls=GROUP_CLS,
)
def bootstrap():
    """
    Boostrap infrastructure and components for Truefoundry platform
    """
    pass


@click.command(
    name="cloud-infra",
    cls=COMMAND_CLS,
    help="Bootstrap Cloud Infrastructure for Truefoundry",
)
@click.option(
    "--dry-run",
    "--dry_run",
    is_flag=True,
    default=False,
)
@click.option(
    "--file",
    is_flag=False,
)
def bootstrap_cloud_infra(dry_run: bool, file: str):
    # If file doesn't exist, then print error
    if file and not os.path.exists(file):
        logger.error(
            f"File at path: {file} not found. Please make sure input file exists"
        )
        return
    infra_controller = InfraController(dry_run=dry_run, config_file_path=file)
    infra_controller.create_infra_using_terraform()


@click.command(
    name="truefoundry",
    cls=COMMAND_CLS,
    help="Bootstrap Truefoundry Agent on an existing kubernetes cluster",
)
@click.option(
    "--dry-run",
    "--dry_run",
    is_flag=True,
    default=False,
)
@click.option(
    "--local",
    is_flag=True,
    default=False,
)
@click.option(
    "--file",
    is_flag=False,
)
def bootstrap_truefoundry(dry_run: bool, file: str, local: bool):
    # If file doesn't exist, then print error
    if not file:
        logger.error("Input file needs to be provided")
        return
    if file and not os.path.exists(file):
        logger.error(
            f"File at path: {file} not found. Please make sure input file exists"
        )
        return
    infra_controller = InfraController(dry_run=dry_run, config_file_path=file)
    infra_controller.install_truefoundry_to_k8s(local)
    return


@click.command(
    name="cluster",
    cls=COMMAND_CLS,
    help="Bootstrap Kubernetes cluster with the control plane",
)
@click.option(
    "--cluster-token",
    "--cluster_token",
    prompt=True,
)
@click.option("--org", prompt=True)
@click.option("--control-plane-url", prompt=True)
def bootstrap_agent_cluster(cluster_token: str, org: str, control_plane_url: str):
    config = {
        "provider": "aws",
        "tenant_name": org,
        "control_plane_url": control_plane_url,
        "tfy_agent": {"cluster_token": cluster_token},
    }
    input_file = f"{org}-input.yaml"
    with open(input_file, "w") as f:
        f.write(json.dumps(config, indent=2))
    infra_controller = InfraController(dry_run=False, config_file_path=input_file)
    infra_controller.install_truefoundry_to_k8s(local=False)
    return


def get_infra_command():
    bootstrap.add_command(bootstrap_cloud_infra)
    bootstrap.add_command(bootstrap_truefoundry)
    bootstrap.add_command(bootstrap_agent_cluster)
    return bootstrap



from typing import Union

from pydantic import Field, conint, constr

from servicefoundry.auto_gen import models
from servicefoundry.lib.model.entity import Deployment
from servicefoundry.v2.lib.deploy import deploy_component


class Service(models.Service):
    type: constr(regex=r"service") = "service"
    resources: models.Resources = Field(default_factory=models.Resources)
    # This is being patched because cue export marks this as a "number"
    replicas: Union[conint(ge=1, le=20), models.Autoscaling] = Field(
        1,
        description="+label=Replicas\n+usage=Replicas of service you want to run\n+icon=fa-clone\n+sort=3",
    )

    class Config:
        extra = "forbid"

    def deploy(self, workspace_fqn: str, wait: bool = True) -> Deployment:
        return deploy_component(component=self, workspace_fqn=workspace_fqn, wait=wait)


class Job(models.Job):
    type: constr(regex=r"job") = "job"
    resources: models.Resources = Field(default_factory=models.Resources)

    class Config:
        extra = "forbid"

    def deploy(self, workspace_fqn: str, wait: bool = True) -> Deployment:
        return deploy_component(component=self, workspace_fqn=workspace_fqn, wait=wait)


class ModelDeployment(models.ModelDeployment):
    type: constr(regex=r"model-deployment") = "model-deployment"
    resources: models.Resources = Field(default_factory=models.Resources)

    class Config:
        extra = "forbid"

    def deploy(self, workspace_fqn: str, wait: bool = True) -> Deployment:
        return deploy_component(component=self, workspace_fqn=workspace_fqn, wait=wait)


class Application(models.Application):
    class Config:
        extra = "forbid"

    def deploy(self, workspace_fqn: str, wait: bool = True) -> Deployment:
        return deploy_component(
            component=self.__root__, workspace_fqn=workspace_fqn, wait=wait
        )



import json
from typing import Dict, Optional, Sequence

import rich_click as click

from servicefoundry.cli.const import COMMAND_CLS, GROUP_CLS
from servicefoundry.cli.util import handle_exception_wrapper
from servicefoundry.lib.dao import application


@click.group(name="trigger", cls=GROUP_CLS)
def trigger_command():
    """
    Trigger a deployed job asynchronously
    """
    pass


@click.command(
    name="job",
    cls=COMMAND_CLS,
    context_settings=dict(ignore_unknown_options=True),
    help="Trigger a Job asynchronously",
)
@click.option(
    "--application-fqn",
    "--application_fqn",
    type=click.STRING,
    required=True,
    help="FQN of the deployment of the Job. This can be found on the Job details page.",
)
@click.option("--command", type=click.STRING, required=False, help="Command to run")
@click.option(
    "--params",
    type=click.UNPROCESSED,
    required=False,
    help="Parameters to pass to the job",
)
@handle_exception_wrapper
def trigger_job(
    application_fqn: str,
    command: Optional[Sequence[str]],
    params: Optional[Dict[str, str]],
):
    if params:
        try:
            params = json.loads(params)
        except json.JSONDecodeError:
            raise Exception("Invalid JSON passed for `params`")
    application.trigger_job(
        application_fqn=application_fqn, command=command, params=params
    )


def get_trigger_command():
    trigger_command.add_command(trigger_job)
    return trigger_command



import os
from pathlib import Path

from servicefoundry.io.rich_output_callback import RichOutputCallBack

DEFAULT_BASE_URL = "https://app.truefoundry.com"
API_SERVER_RELATIVE_PATH = "api/svc"
DEFAULT_API_SERVER = f"{DEFAULT_BASE_URL.rstrip('/')}/{API_SERVER_RELATIVE_PATH}"
DEFAULT_AUTH_UI = DEFAULT_BASE_URL
DEFAULT_AUTH_SERVER = (
    "https://auth-server.tfy-ctl-euwe1-production.production.truefoundry.com"
)
DEFAULT_TENANT_NAME = "truefoundry"
DEFAULT_PROFILE_NAME = "default"
API_KEY_ENV_NAME = "TFY_API_KEY"
HOST_ENV_NAME = "TFY_HOST"

_SFY_CONFIG_DIR = Path.home() / ".truefoundry"
SFY_CONFIG_DIR = str(_SFY_CONFIG_DIR)  # as a directory
OLD_SFY_PROFILES_FILEPATH = _SFY_CONFIG_DIR / "profiles.json"  # as a directory
OLD_SFY_SESSIONS_FILEPATH = _SFY_CONFIG_DIR / "sessions.json"  # as a directory
OLD_SESSION_FILEPATH = str(
    Path.home() / ".truefoundry"
)  # as a filepath, to be removed in future versions
CREDENTIAL_FILEPATH = _SFY_CONFIG_DIR / "credentials.json"

# Build related Config
SERVICE_DEF_FILE_NAME = "servicefoundry.yaml"
TEMPLATE_DEF_FILE_NAME = "template.yaml"
SFY_DIR = ".servicefoundry"
BUILD_DIR = os.path.join(SFY_DIR, "build")

COMPONENT = "Component"
BUILD_PACK = "BuildPack"
KIND = "kind"

# Polling during login redirect
MAX_POLLING_RETRY = 100
POLLING_SLEEP_TIME_IN_SEC = 4

# Refresh access token cutoff
REFRESH_ACCESS_TOKEN_IN_SEC = 10 * 60

ENTITY_JSON_DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"
RICH_OUTPUT_CALLBACK = RichOutputCallBack()

VERSION_PREFIX = "v1"

TERRAFORM_VERSION = "1.3.5"
TERRAGRUNT_VERSION = "0.40.2"
HELM_VERSION = "3.8.0"
KUBECTL_VERSION = "1.26.0"
NSC_VERSION = "2.7.6"

SFY_DEBUG_ENV_KEY = "SFY_DEBUG"
SFY_EXPERIMENTAL_ENV_KEY = "SFY_EXPERIMENTAL"
SFY_INTERNAL_ENV_KEY = "SFY_INTERNAL"



import questionary

from servicefoundry.io.input_hook import InputHook
from servicefoundry.lib.clients.service_foundry_client import (
    ServiceFoundryServiceClient,
)


class NotebookInputHook(InputHook):
    def confirm(self, prompt, default=False):
        raise NotImplementedError()

    def ask_file_path(self, param):
        raise NotImplementedError()

    def ask_python_file(self, param):
        raise NotImplementedError()

    def __init__(self, tfs_client: ServiceFoundryServiceClient):
        super().__init__(tfs_client)

    def ask_string(self, param):
        return questionary.text(param.prompt, default=param.default).ask()

    def ask_number(self, param):
        while True:
            value = questionary.text(param.prompt, default=str(param.default)).ask()
            if value.isdigit():
                return int(value)
            else:
                print("Not an integer Value. Try again")

    def ask_option(self, param):
        return questionary.select(param.prompt, choices=param.options).ask()



import logging

import rich_click as click

from servicefoundry.cli.const import COMMAND_CLS
from servicefoundry.cli.util import handle_exception_wrapper
from servicefoundry.lib.dao import application as application_lib
from servicefoundry.logger import logger


@click.command(name="redeploy", cls=COMMAND_CLS)
@click.option(
    "--application-fqn",
    type=click.STRING,
    default=None,
    help="Application fqn",
    required=True,
)
@click.option(
    "--version",
    type=click.STRING,
    default=None,
    help="Application deployment version",
    required=True,
)
@click.option(
    "--wait/--no-wait",
    is_flag=True,
    show_default=True,
    default=True,
    help="wait and tail the deployment progress",
)
@handle_exception_wrapper
def redeploy_command(application_fqn: str, version: int, wait: bool):
    """
    Redeploy specific version of an application
    """
    _deployment = application_lib.redeploy_application(
        application_fqn=application_fqn, version=version, wait=wait
    )


def get_redeploy_command():
    return redeploy_command



from servicefoundry.auto_gen.models import (
    AppProtocol,
    CanaryStep,
    ConcurrencyPolicy,
    Protocol,
)
from servicefoundry.core import login, logout
from servicefoundry.v2.lib.deployable_patched_models import (
    Application,
    Job,
    ModelDeployment,
    Service,
)
from servicefoundry.v2.lib.patched_models import (
    Autoscaling,
    BasicAuthCreds,
    BlueGreen,
    Build,
    CPUUtilizationMetric,
    CUDAVersion,
    Canary,
    DockerFileBuild,
    FileMount,
    GPU,
    GPUType,
    GitSource,
    HealthProbe,
    HttpProbe,
    HuggingfaceModelHub,
    Image,
    LocalSource,
    Manual,
    Param,
    Port,
    PythonBuild,
    RPSMetric,
    RemoteSource,
    Resources,
    Rolling,
    Schedule,
    TruefoundryModelRegistry,
)
from servicefoundry.version import __version__
from servicefoundry.lib.dao.application import (
    delete_application,
    get_application,
    list_applications,
    trigger_job,
)
from servicefoundry.lib.dao.version import (
    get_version as get_application_version,
    list_versions as list_application_versions,
)
from servicefoundry.lib.dao.workspace import (
    delete_workspace,
    get_workspace_by_fqn,
    list_workspaces,
)



from typing import Optional

from servicefoundry.lib.clients.service_foundry_client import (
    ServiceFoundryServiceClient,
)


def list_versions(
    application_fqn: str,
    client: Optional[ServiceFoundryServiceClient] = None,
):
    client = client or ServiceFoundryServiceClient()
    application = client.get_id_from_fqn(fqn=application_fqn, fqn_type="app")
    versions = client.list_versions(application_id=application["applicationId"])
    return versions


def get_version(
    application_fqn: str,
    version: int,
    client: Optional[ServiceFoundryServiceClient] = None,
):
    client = client or ServiceFoundryServiceClient()
    application = client.get_id_from_fqn(fqn=application_fqn, fqn_type="app")
    versions = client.list_versions(
        application_id=application["applicationId"], deployment_version=version
    )
    if len(versions) == 0:
        raise ValueError(
            f"Version {version!r} for Application with FQN {application_fqn!r} does not exist."
        )
    return versions[0]



# This is meant to hold experimental features - stuff that is provisional and allowed to break
# DO NOT import stuff from here globally. Always import it locally restricted to as smaller scope as possible
# Always guard the imports under servicefoundry.lib.util.is_experimental_env_set
import functools

from servicefoundry.logger import logger


def _warn_on_call(fn):
    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        logger.warning(
            f"Warning: This feature {fn.__name__} is in experimental stage. "
            f"As such there is no guarantees this will be maintained with backward compatibility "
            f"or even available moving forward"
        )
        return fn(*args, **kwargs)

    return wrapper


def trigger_job(*args, **kwargs):
    raise Exception(
        "This function has been stabilised and moved out of experimental module. "
        "Please import it as `from servicefoundry import trigger_job`"
    )



import abc
import inspect
from typing import Dict, Optional

from servicefoundry.function_service.remote.method import method_factory
from servicefoundry.function_service.utils import get_qual_name


class Remote(abc.ABC):
    @property
    @abc.abstractmethod
    def instance(self):
        ...

    @abc.abstractmethod
    def get_qual_name(self, method) -> str:
        ...


class RemoteFunction(Remote):
    def __init__(self, func, **kwargs):
        self._func = func
        self._instance = inspect.getmodule(self._func)
        self.method = method_factory(remote_object=self, method=self._func)

    def __call__(self, *args, **kwargs):
        if len(args) > 0:
            raise Exception(
                "Positional arguments are not supported now.\n"
                "Please use keyword arguments"
            )
        return self.method.__call__(**kwargs)

    def run(self, **kwargs):
        return self.method.run(**kwargs)

    async def run_async(self, *args, **kwargs):
        return await self.method.run_async(*args, **kwargs)

    @property
    def instance(self):
        return self._instance

    def get_qual_name(self, method) -> str:
        return get_qual_name(method)


class RemoteClass(Remote):
    def __init__(
        self,
        class_,
        *,
        init_kwargs: Optional[Dict] = None,
        name: Optional[str] = None,
        **kwargs,
    ):
        self._class = class_
        self._name = name or class_.__name__
        self._init_kwargs = init_kwargs or {}

        self._instance = None

        ###############
        # Maybe this is not the right way to model it.
        # Another approach can be, identify whether the environment is local
        # or our service. If local, then we can just return the instance of
        # the user defined class. Else, we return a class which reflects the
        # methods in the class, but internally doing network call.
        # But I am sticking with this now because not all the concepts of a python
        # class right now, will work in both remote and local. Examples, you cannot
        # access arbitary attributes, properties etc. I want to keep the behaviour
        # identical for now.
        absent = object()
        methods = inspect.getmembers(class_, predicate=inspect.isfunction)
        for method_name, method in methods:
            if method_name.startswith("_"):
                continue
            existing_attribute = getattr(self, method_name, absent)
            if existing_attribute is not absent:
                raise Exception(
                    f"Cannot reflect {method_name!r} of class {class_.__name__!r}.\n"
                    f"{method_name!r} is reserved. Use a different method name"
                )

            setattr(
                self,
                method_name,
                method_factory(remote_object=self, method=method),
            )
        ###############

    def get_qual_name(self, method) -> str:
        return f"{self.name}:{get_qual_name(method)}"

    @property
    def instance(self):
        if self._instance:
            return self._instance
        instance = self._class(**self._init_kwargs)
        self._instance = instance
        return self._instance

    @property
    def class_(self):
        return self._class

    @property
    def name(self) -> str:
        return self._name

    @property
    def init_kwargs(self) -> Dict:
        return self._init_kwargs


def remote(func_or_class, **kwargs):
    if inspect.isfunction(func_or_class):
        return RemoteFunction(func_or_class, **kwargs)
    if inspect.isclass(func_or_class):
        return RemoteClass(func_or_class, **kwargs)

    raise Exception()


if __name__ == "__main__":

    def foo(a, b):
        print(a, b)

    remote(foo)(a=1, b=2)

    class Foo:
        def __init__(self, a, b):
            self.a = a
            self.b = b

        def foo(self):
            print(self.a, self.b)

    foo_instance = remote(Foo, name="foo", init_kwargs={"a": 1, "b": 2})
    foo_instance_2 = remote(Foo, name="foo_2", init_kwargs={"a": 3, "b": 2})

    foo_instance.foo()
    foo_instance_2.foo()
    foo_instance.foo.run()
    # foo_instance.foo.async_run()



import logging

from my_module import Bot

from servicefoundry.function_service import FunctionService

logging.basicConfig(level=logging.INFO)

bot_service = FunctionService(name="bot-service", port=4000)

bot_service.register_class(Bot, init_kwargs={"bot_name": "Alice"}, name="alice_bot")
bot_service.register_class(Bot, init_kwargs={"bot_name": "Bob"}, name="bob_bot")

bot_service.deploy("v1:local:my-ws-2")
# bot_service.run()
#
# print(bot_service)
#
# another_bot_service = FunctionService(name="another_bot_service", port=4001)
# another_bot_service.register_class(
#     Bot, init_kwargs={"bot_name": "Rob"}, name="rob_bot"
# )
# another_bot_service.run().join()



import os
import re
from typing import TypeVar, Union

from servicefoundry.lib.const import (
    SFY_DEBUG_ENV_KEY,
    SFY_EXPERIMENTAL_ENV_KEY,
    SFY_INTERNAL_ENV_KEY,
)


def is_debug_env_set() -> bool:
    return True if os.getenv(SFY_DEBUG_ENV_KEY) else False


def is_experimental_env_set() -> bool:
    # TODO (chiragjn): one of these need to be removed
    return (
        True
        if os.getenv(SFY_EXPERIMENTAL_ENV_KEY) or os.getenv(SFY_INTERNAL_ENV_KEY)
        else False
    )


def get_application_fqn_from_deployment_fqn(deployment_fqn: str) -> str:
    if not re.search(r":\d+$", deployment_fqn):
        raise ValueError(
            "Invalid `deployment_fqn` format. A deployment fqn is supposed to end with a version number"
        )
    application_fqn, _ = deployment_fqn.rsplit(":", 1)
    return application_fqn


def get_deployment_fqn_from_application_fqn(
    application_fqn: str, version: Union[str, int]
) -> str:
    return f"{application_fqn}:{version}"



import subprocess

from servicefoundry.logger import logger


class Shell:
    def execute_shell_command(self, command, ip=None):
        logger.debug(f"executing command: {command}")
        try:
            p = subprocess.run(command, stdout=subprocess.PIPE, check=True, input=ip)
            return p.stdout.decode("UTF-8")
        except subprocess.CalledProcessError as err:
            raise Exception(f"failed to execute: {command}, error: {err}")



import sys
import time
from typing import List, Optional, TypeVar

from pydantic import BaseModel
from rich.status import Status

from servicefoundry.auto_gen import models as auto_gen_models
from servicefoundry.builder.docker_service import env_has_docker
from servicefoundry.lib.clients.service_foundry_client import (
    ServiceFoundryServiceClient,
)
from servicefoundry.lib.clients.utils import poll_for_function
from servicefoundry.lib.dao.workspace import get_workspace_by_fqn
from servicefoundry.lib.model.entity import Deployment, DeploymentTransitionStatus
from servicefoundry.lib.util import get_application_fqn_from_deployment_fqn
from servicefoundry.logger import logger
from servicefoundry.v2.lib.models import BuildResponse
from servicefoundry.v2.lib.source import (
    local_source_to_image,
    local_source_to_remote_source,
)

Component = TypeVar("Component", bound=BaseModel)


def _handle_if_local_source(component: Component, workspace_fqn: str) -> Component:
    if (
        hasattr(component, "image")
        and isinstance(component.image, auto_gen_models.Build)
        and isinstance(component.image.build_source, auto_gen_models.LocalSource)
    ):
        new_component = component.copy(deep=True)

        if component.image.build_source.local_build:
            if not env_has_docker():
                logger.warning(
                    "Did not find Docker locally installed on this system, image will be built remotely. "
                    "For faster builds it is recommended to install Docker locally. "
                    "If you always want to build remotely, "
                    "please set `image.build_source.local_build` to `false` in your YAML spec or equivalently set "
                    "`image=Build(build_source=LocalSource(local_build=False, ...))` in your "
                    "`Service` or `Job` definition code."
                )
                local_build = False
            else:
                logger.info(
                    "Found locally installed docker, image will be built locally and then pushed. "
                    "If you want to always build remotely instead of locally, "
                    "please set `image.build_source.local_build` to `false` in your YAML spec or equivalently set "
                    "`image=Build(build_source=LocalSource(local_build=False, ...))` in your "
                    "`Service` or `Job` definition code."
                )
                local_build = True
        else:
            logger.info(
                "Image will be built remotely because `image.build_source.local_build` is set to `false`. "
                "For faster builds it is recommended to install Docker locally and "
                "set `image.build_source.local_build` to `true` in your YAML spec "
                "or equivalently set `image=Build(build_source=LocalSource(local_build=True, ...))` "
                "in your `Service` or `Job` definition code."
            )
            local_build = False

        if local_build:
            # We are to build the image locally, push and update `image` in spec
            logger.info("Building image for %s '%s'", component.type, component.name)
            new_component.image = local_source_to_image(
                build=component.image,
                docker_registry_fqn=component.image.docker_registry,
                workspace_fqn=workspace_fqn,
                component_name=component.name,
            )
        else:
            # We'll build image on Truefoundry servers, upload the source and update image.build_source
            logger.info("Uploading code for %s '%s'", component.type, component.name)
            new_component.image.build_source = local_source_to_remote_source(
                local_source=component.image.build_source,
                workspace_fqn=workspace_fqn,
                component_name=component.name,
            )
            logger.debug("Uploaded code for %s '%s'", component.type, component.name)
        return new_component
    return component


def _log_application_dashboard_url(deployment: Deployment, log_message: str):
    application_id = deployment.applicationId

    # TODO: is there any simpler way to get this? :cry
    client = ServiceFoundryServiceClient()

    url = f"{client.base_url.strip('/')}/applications/{application_id}?tab=deployments"
    logger.info(log_message, url)


def _tail_build_logs(build_responses: List[BuildResponse]) -> None:
    client = ServiceFoundryServiceClient()

    # TODO: Explore other options like,
    # https://rich.readthedocs.io/en/stable/live.html#live-display
    # How does docker/compose does multiple build logs?
    for build_response in build_responses:
        logger.info("Tailing build logs for '%s'", build_response.componentName)
        client.tail_build_logs(build_response=build_response, wait=True)


def _deploy_wait_handler(
    deployment: Deployment,
) -> Optional[DeploymentTransitionStatus]:
    _log_application_dashboard_url(
        deployment=deployment,
        log_message=(
            "You can track the progress below or on the dashboard:- '%s'\n"
            "You can press Ctrl + C to exit the tailing of build logs "
            "and deployment will continue on the server"
        ),
    )
    with Status(status="Polling for deployment status") as spinner:
        last_status_printed = None
        client = ServiceFoundryServiceClient()
        start_time = time.monotonic()
        total_timeout_time: int = 300
        poll_interval_seconds = 5
        time_elapsed = 0

        for deployment_statuses in poll_for_function(
            client.get_deployment_statuses,
            poll_after_secs=poll_interval_seconds,
            application_id=deployment.applicationId,
            deployment_id=deployment.id,
        ):
            if len(deployment_statuses) == 0:
                logger.warning("Did not receive any deployment status")
                continue

            latest_deployment_status = deployment_statuses[-1]

            status_to_print = (
                latest_deployment_status.transition or latest_deployment_status.status
            )
            spinner.update(status=f"Current state: {status_to_print}")
            if status_to_print != last_status_printed:
                if DeploymentTransitionStatus.is_failure_state(status_to_print):
                    logger.error("State: %r", status_to_print)
                else:
                    logger.info("State: %r", status_to_print)
                last_status_printed = status_to_print

            if latest_deployment_status.state.isTerminalState:
                break

            if (
                latest_deployment_status.transition
                == DeploymentTransitionStatus.BUILDING
            ):
                build_responses = client.get_deployment_build_response(
                    application_id=deployment.applicationId, deployment_id=deployment.id
                )
                _tail_build_logs(build_responses)

            time_elapsed = time.monotonic() - start_time
            if time_elapsed > total_timeout_time:
                logger.warning("Polled server for %s secs.", int(time_elapsed))
                break

    return last_status_printed


def _warn_when_gpu_selected_without_cuda(component: Component):
    if (
        hasattr(component, "resources")
        and hasattr(component, "image")
        and component.resources
        and component.resources.gpu
        and isinstance(component.image, auto_gen_models.Build)
        and isinstance(component.image.build_spec, auto_gen_models.PythonBuild)
        and not component.image.build_spec.cuda_version
    ):
        logger.warning(
            "Warning: `gpu` is set in `Resources` (i.e. `resources.gpu`) "
            "but no `cuda_version` was passed to `PythonBuild` "
            "(i.e. `image.build_spec.cuda_version`). "
            "Your application might optionally need CUDA toolkit installed "
            "to utilize the GPU. You can choose one by passing one of "
            "`servicefoundry.CUDAVersion` in `PythonBuild` instance. "
            "\n\nE.g.\n```\nPythonBuild(..., cuda_version=CUDAVersion.CUDA_11_3_CUDNN8)\n```"
        )


def deploy_component(
    component: Component, workspace_fqn: str, wait: bool = True
) -> Deployment:
    _warn_when_gpu_selected_without_cuda(component=component)
    workspace_id = get_workspace_by_fqn(workspace_fqn).id
    updated_component = _handle_if_local_source(
        component=component, workspace_fqn=workspace_fqn
    )
    client = ServiceFoundryServiceClient()
    response = client.deploy_application(
        workspace_id=workspace_id, application=updated_component
    )
    logger.info(
        "🚀 Deployment started for application '%s'. Deployment FQN is '%s'.",
        updated_component.name,
        response.fqn,
    )
    if wait:
        try:
            last_status_printed = _deploy_wait_handler(deployment=response)
            if not last_status_printed or DeploymentTransitionStatus.is_failure_state(
                last_status_printed
            ):
                deployment_tab_url = f"{client.base_url.strip('/')}/applications/{response.applicationId}?tab=deployments"
                message = f"Deployment Failed. Please refer to the logs for additional details - {deployment_tab_url}"
                sys.exit(message)
        except KeyboardInterrupt:
            logger.info("Ctrl-c executed. The deployment will still continue.")

    deployment_fqn = response.fqn
    application_fqn = get_application_fqn_from_deployment_fqn(deployment_fqn)
    logger.info("Deployment FQN: %s", deployment_fqn)
    logger.info("Application FQN: %s", application_fqn)

    _log_application_dashboard_url(
        deployment=response,
        log_message="You can find the application on the dashboard:- '%s'",
    )
    return response



import sys


class OutputCallBack:
    def print_header(self, line):
        sys.stdout.write(f"{line}\n")

    def print_line(self, line):
        sys.stdout.write(f"{line}\n")

    def print_lines_in_panel(self, lines, header=None):
        self.print_header(header)
        for line in lines:
            self.print_line(line)

    def print_code_in_panel(self, lines, header=None, lang="python"):
        self.print_lines_in_panel(lines, header)

    def print(self, line):
        # just an alias
        self.print_line(line)
